﻿using System;
using System.IO;
using System.Net;
using System.Web;
using System.Text;
using System.Threading;
using Plywood.Compiler;
using Plywood;
using System.Collections;
using System.Diagnostics;
using System.Reflection;
using System.Windows.Forms;
using System.Runtime.InteropServices;

static class CommandSystem {
    private static FileSystem p_FileSystem;
    private static bool p_WriteLog = false;
    private static ITerminal p_MainTerminal;
    private static RunState p_State;

    [DllImport("user32.dll")]
    private static extern void ShowWindow(IntPtr hwnd, int state);

    public static void InitSubSystem() {
        //Close the previous file system?
        if (p_FileSystem != null) {
            try { p_FileSystem.Close(); }
            catch { }
        }

        //Define the output terminal to windows console
        if (p_MainTerminal == null) {
            p_MainTerminal = new WindowsTerminal();
        }
        //Define the file system
        p_FileSystem = new FileSystem("./fileSystem.txt");
    }
    
    public static ITerminal Terminal { get { return p_MainTerminal; } set { p_MainTerminal = value; } }

    public static void Print(string line) {
        //Just print to console
        p_MainTerminal.MessageSent(p_State, TerminalMessageType.Write, line + "\n", false);

        //Write?
        if (p_WriteLog) {
            FileStream fs = new FileStream("./log.log", FileMode.OpenOrCreate);
            byte[] w = System.Text.Encoding.ASCII.GetBytes(line + Environment.NewLine);
            fs.Seek(0, SeekOrigin.End);
            fs.Write(w, 0, w.Length);
            fs.Close();
        }
    }
    public static void PrintError(string line) { 
        //Just print the line
        Print("Error: " + line);
    }

    private static bool grabBoolFromArgs(string args) { 
        //Make the argument lower case 
        args = args.ToLower();

        //Is it true/false?
        if (args == "true" || args == "1") { return true; }
        else if (args == "false" || args == "0") { return false; }
        else { 
            //Print error
            PrintError("Invalid arguments");
            return false;
        }
    }
}