﻿using System;
using System.IO;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Runtime.InteropServices;
using System.Reflection;

namespace Plywood {
    public static class DynamicLibraryLoader {
        [DllImport("kernel32.dll")]
        private static extern IntPtr LoadLibraryEx(string filename, IntPtr hFile, uint dwFlags);

        [DllImport("kernel32.dll")]
        private static extern bool FreeLibrary(IntPtr dllPointer);

        [DllImport("kernel32.dll")]
        private static extern IntPtr GetProcAddress(IntPtr dllPointer, string functionName);

        private static IntPtr LoadWin32(string filename) { 
            //Get the initial handle
            IntPtr buffer = LoadLibraryEx(filename, IntPtr.Zero, 0x00000008);

            //Check
            if (buffer == IntPtr.Zero) {
                throw new Exception("Cannot load " + filename);
            }

            //Return the handle
            return buffer;
        }

        public static object InvokeMethod(string filename, string method, object[] args, Type returnType) {
            //Invoke the method
            return getMethod(filename, method, args, returnType).DynamicInvoke(args);
        }
        public static void AddMethod(string filename, string method, object[] args, Type returnType) {
            //Let the local method handle this
            getMethod(filename, method, args, returnType);
        }
        public static Delegate GetDynamicDelegate(Type returnType, Type[] arguments, out Type CallbackDelegate, out FieldInfo CallbackField) {
            //Define the c# script to compile  
            string argCode = "";
            for (int c = 0; c < arguments.Length; c++) {
                argCode += getCSType(arguments[c]) + " a" + (c.ToString()) + ",";
            }
            if (arguments.Length != 0) { argCode = argCode.Substring(0, argCode.Length - 1); }
            string code = "using Plywood;public delegate " + getCSType(returnType) + " DelegateHandler(" + argCode + ",IMethod method);";
            code += "public delegate object DelegateCallHandler(object[] param,IMethod method);";
            code += "public static class DelegateClass{";
            code += "public static DelegateCallHandler CallHandler;";
            code += "public static " + getCSType(returnType) + " DelegateMethod(" + argCode + ", IMethod method){";
            code += "object[]p={";
            for (int c = 0; c < arguments.Length; c++) {
                code += "a" + c.ToString() + ",";
            }
            if (arguments.Length != 0) { code = code.Substring(0, code.Length - 1); }
            code += "};";
            code += (returnType != null ? "return (" + getCSType(returnType) + ")" : "") + "CallHandler(p,method);}}";

            //Compile it
            CompilerResults results;
            CompilerParameters param = new CompilerParameters();
            CodeDomProvider provider = Microsoft.CSharp.CSharpCodeProvider.CreateProvider("cs");
            ICodeCompiler compiler = provider.CreateCompiler();
            param.GenerateInMemory = true;
            param.GenerateExecutable = false;
            param.ReferencedAssemblies.Add(new FileInfo(Environment.SystemDirectory + "..\\..\\" + "Microsoft.NET\\Framework\\v2.0.50727\\System.dll").FullName);
            param.ReferencedAssemblies.Add(typeof(DynamicLibraryLoader).Assembly.CodeBase.Replace("file:///", ""));
            results = compiler.CompileAssemblyFromSource(param, code); 
            
            //Get the delegate
            Type delType = results.CompiledAssembly.GetType("DelegateHandler");
            MethodInfo method = results.CompiledAssembly.GetType("DelegateClass").GetMethod("DelegateMethod");
            CallbackDelegate = results.CompiledAssembly.GetType("DelegateCallHandler");
            CallbackField = results.CompiledAssembly.GetType("DelegateClass").GetField("CallHandler");
            return Delegate.CreateDelegate(delType, method);
        }
        
        private static Delegate getMethod(string filename, string method, object[] args, Type returnType) {
            //Load the library
            IntPtr wHandle = LoadWin32(filename);
            IntPtr methodPointer = GetProcAddress(wHandle, method);

            //Check the method
            if (methodPointer == IntPtr.Zero) {
                throw new Exception("Method \"" + method + "\" was not found");
            }

            //Get the types of the arguments
            Type[] types = new Type[args.Length];
            for (int c = 0; c < types.Length; c++) { types[c] = args[c].GetType(); }

            //Define the dynamic delegate
            Type dynamicDel = null;
            
            //Is it in cache?
            bool isCache =false;
            foreach (delegateTypeCache cache in p_Cache) {
                if (!(cache.Method == method && 
                      filename.ToLower() == cache.Filename.ToLower() &&
                      cache.ReturnType == returnType)) { continue; }
                int e = 0;
                if (cache.Arguments.Length != types.Length) { continue; }
                for (int c = 0; c < cache.Arguments.Length; c++) {
                    if (cache.Arguments[c] != types[c]) { break; ; }
                    e++;
                }
                if (e == cache.Arguments.Length) {
                    dynamicDel = cache.Type;                    
                    isCache = true;
                }
            }

            if (!isCache) { 
                //Create it
                dynamicDel = createDynamicDelegate(returnType, types); 
                
                //Add it to cache
                while (p_Hang) ;
                p_Hang = true;
                Array.Resize(ref p_Cache, p_Cache.Length + 1);
                p_Cache[p_Cache.Length - 1] = new delegateTypeCache() { 
                    Arguments = types,
                    Filename = filename, 
                    Method = method,
                    Type = dynamicDel,
                    ReturnType = returnType
                };
                p_Hang = false;
            }

            //Create the delegate for the method pointer
            return Marshal.GetDelegateForFunctionPointer(methodPointer, dynamicDel);
        }
        private static Type createDynamicDelegate(Type returnType, Type[] types) { 
            //Define the c# code
            string csCode = "using System;public delegate " + getCSType(returnType) + " DYNAMIC_DELEGATE(";
            for(int c = 0; c < types.Length; c++) {
                csCode += getCSType(types[c]) + " a" + c.ToString() + ", ";
            }
            if (types.Length != 0) { csCode = csCode.Substring(0, csCode.Length - 2); }
            csCode += ");";

            //Compile it
            CodeDomProvider provider = Microsoft.CSharp.CSharpCodeProvider.CreateProvider("cs");
            CompilerParameters param = new CompilerParameters();
            CompilerResults results;
            ICodeCompiler compiler = provider.CreateCompiler();
            param.GenerateInMemory = true;
            param.GenerateExecutable = false;
            param.ReferencedAssemblies.Add(new FileInfo(Environment.SystemDirectory + "..\\..\\" +
                                            "Microsoft.NET\\Framework\\v2.0.50727\\System.dll").FullName);
            results = compiler.CompileAssemblyFromSource(param, csCode);
            
            //Grab the delegate
            Type type = results.CompiledAssembly.GetType("DYNAMIC_DELEGATE");

            //Clean up
            results = null;
            provider.Dispose();
            compiler = null;
            csCode = null;
            return type;
        }
        internal static string getCSType(Type t) {
            //Void?
            if (t == typeof(void) || t == null) {
                return "void";
            }

            //Just return the type full name
            return t.FullName.Replace("+", ".");
        }

        private class delegateTypeCache {
            public Type Type;
            public Type ReturnType;
            public string Method;
            public string Filename;
            public Type[] Arguments;
        }
        private static bool p_Hang = false;
        private static delegateTypeCache[] p_Cache = new delegateTypeCache[0];
    }
}