﻿using System;
using System.Collections;
using Plywood.Compiler;

namespace Plywood {
    public partial class GenericMethods {
        public long array() {
            //Define the initial array
            p_State.CurrentDefinition.ArrayList = (IList)new object[0];
            p_State.ApplyDefinitionGlobal = true;
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";

            //Push the object array defined to the current definition
            object[] objects = (object[])p_State.CurrentDefinition.ArrayList;

            //Shall we read values into the array?
            if (!p_State.Script.hasScope(p_State.Stream)) { return 0; }

            //Read the scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Loop through the scope
            string build = "";
            for (int c = 0; c < scope.Length; c++) {
                //Ignore?
                if (scope[c] == '\n') { continue; }

                //Skip braces?
                if (scope[c] == '(') {
                    string b = p_State.Script.getScopeEndIndexWithStr(scope, c, '(', ')');
                    build += "(" + b + ")";
                    c += b.Length + 1;
                    continue;
                }

                //Skip string?
                if (scope[c] == '"') {
                    build += "\"" + p_State.Script.skipString(scope, ref c) + "\"";
                    continue;
                }

                //Entry split character?
                if (scope[c] == ',') {
                    //Add the entry to the array
                    Array.Resize(ref objects, objects.Length + 1);
                    objects[objects.Length - 1] = p_State.Script.getValueFromRaw(build, p_State);
                    build = "";
                    continue;
                }
                build += scope[c];
            }

            //Clean up
            if (build.Length != 0) {
                Array.Resize(ref objects, objects.Length + 1);
                objects[objects.Length - 1] = p_State.Script.getValueFromRaw(build, p_State);
            }

            //Push the object array back into the definition
            p_State.CurrentDefinition.ArrayList = (IList)objects;
            return 0;
        }
        public long array(params object[] values) {
            //Push the array to the current definition
            p_State.CurrentDefinition.ArrayList = (IList)values;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public void array_iterate(string definition, string defValue, string callback) { 
            //Define the definition
            defValue = defValue.Replace(" ", "");
            try {
                if (!p_State.Script.validName(defValue)) { throw new Exception(""); }
                foreach (IDefinition d in p_State.Script.p_Defines) {
                    if (d.Name == defValue) {
                        throw new Exception("");
                    }
                }
            }
            catch {
                throw new Exception("Definition \"" + defValue + "\" already exists");
            }
            Array.Resize(ref p_State.Script.p_Defines, p_State.Script.p_Defines.Length + 1);
            p_State.Script.p_Defines[p_State.Script.p_Defines.Length - 1] = new Interpreter.definition { 
                Name = defValue,
                Value = "0",
                IsNumeric = false
            };

            //Get the callback code
            callback = (string)p_State.Script.getValueFromRaw(callback, p_State);

            //Grab the array
            object[] array = (object[])p_State.Script.getDefByName(definition).ArrayList;

            //Iterate through the array
            foreach (object o in array) { 
                //Set the definition
                for(int c = 0; c < p_State.Script.p_Defines.Length; c++) {
                    if (p_State.Script.p_Defines[c].Name == defValue) { p_State.Script.p_Defines[c].Value = o.ToString(); }
                }

                //Invoke the callback
                execute(callback);
            }
        }
        public void array_set(string definition, long index, string value) {
            //Get the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Set the entry
            def.ArrayList[(int)index] = p_State.Script.getValueFromRaw(value, p_State);

            //Push the definition back into the definitions
            p_State.Script.applyDef(def);
        }
        public object array_index(string value, string index) {
            //Get the index
            int ind = Convert.ToInt32(p_State.Script.getValueFromRaw(index, p_State));

            //Look for the definition
            foreach (IDefinition def in p_State.Script.p_Defines) {
                if (def.Name == value) {
                    if (def.ArrayList == null) { throw new Exception("Definition is not an array"); }
                    return def.ArrayList[(int)ind];
                }
            }

            //Not found
            throw new Exception("Not found");
        }
        public void array_add(string definition, string value) {
            //Get the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Push the array 
            object[] array = (object[])def.ArrayList;

            //Add the entry
            Array.Resize(ref array, array.Length + 1);
            array[array.Length - 1] = p_State.Script.getValueFromRaw(value, p_State);

            //Pop the definition back to the script instance
            def.ArrayList = (IList)array;
            p_State.Script.applyDef(def);
        }
        public void array_addAt(string definition, string i, string value) { 
            //Grab the definition and index
            IDefinition def = p_State.Script.getDefByName(definition);
            int index = Convert.ToInt32(p_State.Script.getValueFromRaw(i, p_State));

            //Get the value
            object val = p_State.Script.getValueFromRaw(value, p_State);

            //Grab the array
            object[] array = (object[])def.ArrayList;

            //Rebuild the array
            object[] rebuild = new object[array.Length + 1];

            //Add the left and right side of the array
            for (int c = 0; c < index; c++) { rebuild[c] = array[c]; }
            for (int c = index; c < array.Length; c++) { rebuild[c + 1] = array[c]; }
            
            //Insert the value
            rebuild[index] = val;

            //Push the array back to the definition
            def.ArrayList = (IList)rebuild;
            p_State.Script.applyDef(def);
        }
        public void array_addArray(string array1, string array2) {
            //Grab both array definition
            IDefinition arr1 = p_State.Script.getDefByName(array1);
            IDefinition arr2 = p_State.Script.getDefByName(array2);

            //Grab the object arrays from each                
            object[] source = (object[])arr1.ArrayList;
            object[] data = (object[])arr2.ArrayList;
            long arr1Length = source.Length;
            //Add the data from data to source
            Array.Resize(ref source, source.Length + data.Length);
            for (int c = 0; c < data.Length; c++) {
                try { source[arr1Length + c] = data[c]; }
                catch { }
            }

            //Push the array back into the definition
            arr1.ArrayList = source;
            p_State.Script.applyDef(arr1);
        }
        public void array_remove(string definition, string value) {
            //Grab the definiion
            IDefinition def = p_State.Script.getDefByName(definition);

            //Get the object instance of the value
            object objInstance = p_State.Script.getValueFromRaw(value, p_State);

            //Get the object array
            object[] array = (object[])def.ArrayList;

            //Look for the value
            int offset = 0;
            foreach (object o in array) {
                if (o.Equals(objInstance)) { offset++; }
            }

            //Was there an entry found?
            if (offset == 0) {
                throw new Exception("Entry does not exist");
            }

            //Define the new array buffer
            object[] buffer = new object[array.Length - offset];
            int ai = 0;

            //Fill in the buffer
            for (int c = 0; c < array.Length; c++) {
                if (array[c].Equals(objInstance)) { continue; }
                buffer[ai] = array[c];
                ai++;
            }

            //Push the bufer to the definition
            def.ArrayList = (IList)buffer;
            p_State.Script.applyDef(def);
        }
        public void array_removeIndex(string definition, string value) {
            //Grab the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Grab the object array
            object[] array = (object[])def.ArrayList;

            //Get the index
            long index = Convert.ToInt64(p_State.Script.getValueFromRaw(value, p_State));

            //Is the index valud?
            if (index >= array.Length || index < 0) {
                throw new Exception("Invalid index");
            }

            //Define the new buffer
            object[] buffer = new object[array.Length - 1];
            int ai = 0;

            //Ignore the specified index
            for (int c = 0; c < array.Length; c++) {
                if (c == index) { continue; }
                buffer[ai] = array[c];
                ai++;
            }

            //Push the new array to the definition
            def.ArrayList = (IList)buffer;
            p_State.Script.applyDef(def);
        }
        public void array_resize(string definition, string size) {
            //Grab the size
            long length = (long)p_State.Script.getValueFromRaw(size, p_State);

            //Grab the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Resize the array
            object[] array = (object[])def.ArrayList;
            Array.Resize(ref array, (int)length);
            def.ArrayList = (IList)array;

            //Apply the definition
            p_State.Script.applyDef(def);
        }
        public void array_sort(string definition) {
            //Get the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Get the array
            object[] array = (object[])def.ArrayList;

            //Convert the array to a list to sort
            double[] buffer = new double[array.Length];
            for (int c = 0; c < array.Length; c++) {
                //Check the type of the array entity
                if (array[c].GetType() != typeof(long) && array[c].GetType() != typeof(double)) {
                    throw new Exception("Sort only works for integer value types");
                }

                //Push the value into the buffer
                buffer[c] = Convert.ToDouble(array[c]);
            }

            //Sort the array
            sort(ref buffer);

            //Convert the buffer to an object array
            array = new object[array.Length];
            for (int c = 0; c < array.Length; c++) {
                array[c] = buffer[c];
            }

            //Push the array back to the definition
            def.ArrayList = (IList)array;
            p_State.Script.applyDef(def);
        }
        public long array_length(string definition) {
            //Return the array length
            return p_State.Script.getDefByName(definition).ArrayList.Count;
        }
        public void array_clear(string definition) {
            //Just set the array length to 0
            array_resize(definition, "0");
        }
        public bool array_isArray(string definition) {
            //Return if the array with this definition is not null
            return p_State.Script.getDefByName(definition).ArrayList != null;
        }
        public void array_invert(string definition) {
            //Grab the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Grab the object array
            object[] array = (object[])def.ArrayList;

            //Invert it
            object[] invert = new object[array.Length];
            for (int c = 0; c < array.Length; c++) {
                invert[c] = array[array.Length - 1 - c];
            }

            //Push the inverted array back to the definition
            def.ArrayList = (IList)invert;
            p_State.Script.applyDef(def);
        }
        private void sort(ref double[] stack) {
            //Worth processing?
            if (stack.Length <= 1) { return; }

            //Grab the pivot value
            double pVal = stack[stack.Length / 2];

            //Define the left and right sides of the pivot
            double[] left = new double[0];
            double[] right = new double[0];

            //Loop through the stack
            for (long c = 0; c < stack.Length; c++) {
                //Pivot index?
                if (c == stack.Length / 2) { continue; }

                //Is this value less than the pivot?
                if (stack[c] < pVal) {
                    Array.Resize(ref left, left.Length + 1);
                    left[left.Length - 1] = stack[c];
                    continue;
                }

                //Add to the right side
                Array.Resize(ref right, right.Length + 1);
                right[right.Length - 1] = stack[c];
            }

            //Sort the left and right values
            sort(ref left);
            sort(ref right);

            //Reposition the values on the stack
            stack[left.Length] = pVal;
            for (long c = 0; c < left.Length; c++) {
                stack[c] = left[c];
            }
            for (long c = 0; c < right.Length; c++) {
                stack[c + left.Length + 1] = right[c];
            }
        }

        public partial class MethodCodes {
            public string array_sort(CodeBuilder builder, string args, bool inline) { 
                //Does the class for auto-gen code exist?
                CodeBuilder.Class cls = builder.GetClass("AUTO_GEN");
           
                //Add the array_sort function?
                if (!cls.MethodExists("array_sort")) { 
                    //Define the method to add
                    CodeBuilder.Method sort = new CodeBuilder.Method("array_sort", null, new Type[] { 
                        typeof(object[])
                    });
                    sort.Accessor = CodeBuilder.Accessor.Public | CodeBuilder.Accessor.Static;
                   
                    /*
                     * Define the code to add
                    */
                                        
                    //Worth processing code
                    sort.AddLine("if(arg0.Length<=1){return;}");

                    //Defining the values to be used in sort
                    sort.AddLine("object[]left=new object[0];object[]right=new object[0];");
                    sort.AddLine("double pVal=Convert.ToDouble(arg0[arg0.Length/2]);");

                    //Getting the left and right side of the pivot
                    sort.AddLine("for(int c=0;c<arg0.Length;c++){");
                    sort.AddLine("if(c==arg0.Length/2){continue;}");
                    sort.AddLine("if(Convert.ToDouble(arg0[c])<pVal){Array.Resize(ref left,left.Length+1);left[left.Length-1]=arg0[c];continue;}");
                    sort.AddLine("Array.Resize(ref right,right.Length+1);right[right.Length-1]=arg0[c];");
                    sort.AddLine("}");

                    //Sort left and right indexes
                    sort.AddLine("array_sort(left);");
                    sort.AddLine("array_sort(right);");

                    //Apply the sorted stacks to the array
                    sort.AddLine("arg0[left.Length]=pVal;");
                    sort.AddLine("for(int c=0;c<left.Length;c++){arg0[c]=left[c];}");
                    sort.AddLine("for(int c=0;c<right.Length;c++){arg0[c+left.Length+1]=right[c];}");
                    
                    //Add the method
                    cls.AddMethod(sort);
                }
                
                //Define the code to invoke the array sort
                string code = "AUTO_GEN.array_sort(" + args + ");";
                builder.SelectedMethod.AddLine(code);
                return code;
            }
        }
    }
}