﻿using System;
using System.Reflection;
using Plywood.Compiler;

namespace Plywood {
    public partial class GenericMethods {
        private RunState p_State;
        private static string[] p_IgnoreLogicalPass = new string[0];

        public GenericMethods(RunState state) { 
            //Define the globals
            p_State = state;
        }

        private static bool ignoreMethod(MethodInfo m) { 
            //Ignore it's name?
            if (m.Name == "ToString" || m.Name == "GetType" || m.Name == "GetHashCode") {
                return true;
            }

            //Don't ignore
            return false;
        }
       
        public IMethod[] GetClrMethods() { 
            //Let the local method handle this
            return GetTypeMethods(typeof(GenericMethods), this);
        }       
        public IMethod[] GetTypeMethods(Type type, object instance) {
            //Grab the methods
            MethodInfo[] methods = type.GetMethods();

            //Define the return buffer
            IMethod[] buffer = new IMethod[methods.Length];
            int nullCount = 0;
            
            //Loop through the type methods
            for (int c = 0; c < methods.Length; c++) {
                if (ignoreMethod(methods[c])) { nullCount++; continue; }
                if (instance.GetType() == typeof(GenericMethods)) {
                    buffer[c] = new GenericMethod(methods[c]);
                    continue;
                }
                buffer[c] = new ClrMethod(methods[c]) { AssignedInstance = instance };
            }

            //Rebuild null references
            IMethod[] rebuild = new IMethod[buffer.Length - nullCount];
            int ai = 0;
            for (int c = 0; c < buffer.Length; c++) {
                if (buffer[c] == null) { continue; }
                rebuild[ai] = buffer[c];
                ai++;
            }


            //Return the buffer
            return rebuild;
        }

        private object getPropValue(string value, PropertyInfo prop) {
            //Is it bool?
            if (prop.PropertyType == typeof(bool)) { return Convert.ToBoolean(value); }
            //Is it byte?
            if (prop.PropertyType == typeof(byte)) { return Convert.ToByte(value); }
            //Is it string?
            if (prop.PropertyType == typeof(string)) { return value; }
            //Is it integer?
            if (prop.PropertyType == typeof(int)) { return Convert.ToInt32(value); }
            //Is it enum?
            if (prop.PropertyType.IsEnum) {
                //Return the appropriate enumeriation field
                return Enum.Parse(prop.PropertyType, value);
            }

            //Not found
            throw new Exception("Not found");
        }
        private string applyValueToType(object instance, string value, string variable, Type type) {
            //Find the property
            foreach (PropertyInfo p in type.GetProperties()) {
                if (p.Name != variable) { continue; }

                //Shall we set the value?
                if (value != null) {
                    p.SetValue(instance, getPropValue(value, p), null);
                }

                //Return the value
                object vObj = p.GetValue(instance, null);
                if (p.PropertyType == typeof(bool)) { return vObj.ToString().ToLower(); }
                if (vObj == null) { return ""; }
                return vObj.ToString();
            }

            //Not found
            throw new Exception("Variable not found");
        }
       
        public static bool IgnoreLogicPass(string functionName) {
            //Defined to ignore?
            foreach (string s in p_IgnoreLogicalPass) {
                if (functionName == s) { return true; }
            }

            //Define the list of methods which require the logic statement passed
            string[] fList = { 
                                    //Syntactical functions
                                    "if",
                                    "while",
                                    "for",
                                    "foreach",
                                    "destruct",
                                    
                                    //Array functions
                                    "array_iterate",
                                    "array_isArray",
                                    "array_set",
                                    "array_index",
                                    "array_add",
                                    "array_addAt",
                                    "array_addArray",
                                    "array_remove",
                                    "array_removeIndex",
                                    "array_resize",
                                    "array_sort",                                    
                                    "array_sortBubble",
                                    "array_length",
                                    "array_clear",
                                    "array_invert",
                                    
                                    //Math functions
                                    "max", 
                                    "min",
                                    "average",

                                    //Network functions
                                    "net_socketSendBytes",

                                    //Common language runtime functions
                                    "clr_getValue",
                                    "clr_setValue",
                                    "clr_createHandle",
                                    "clr_toString",

                                    //IO functions
                                    "io_writeArray",

                                    //Management
                                    "manage_removeStack"
                                 };

            //Does the function name passed fall in this list?
            foreach (string s in fList) {
                if (s == functionName) { return true; }
            }

            //None found
            return false;
        }
        public static void AddIgnoreLogicalPass(string functionName) { 
            //Already ignore?
            if (IgnoreLogicPass(functionName)) { return; }

            //Add it
            Array.Resize(ref p_IgnoreLogicalPass, p_IgnoreLogicalPass.Length + 1);
            p_IgnoreLogicalPass[p_IgnoreLogicalPass.Length - 1] = functionName;
        }
    }
}