﻿using System;
using System.IO;
using System.Reflection;
using System.Collections;

namespace Plywood {
    public partial class GenericMethods {
        private static clrObj[] p_ClrObjects = new clrObj[0];
        private static string[] p_ClrImports = new string[0];

        private object[] clrConvMethodTypes(object[] args, ParameterInfo[] types) { 
            //Let the local method handle this
            Type[] t = new Type[types.Length];
            for (int c = 0; c < t.Length; c++) { t[c] = types[c].ParameterType; }
            return clrConvMethodTypes(args, t);
        }
        private object[] clrConvMethodTypes(object[] args, Type[] types) {
            //Check
            if (args.Length != types.Length) {
                throw new Exception("Invalid arguments");
            }

            //Define the return buffer
            object[] buffer = new object[args.Length];

            //Loop through the arguments
            for (int c = 0; c < args.Length; c++) { 
                //Get the appropriate type
                Type t = types[c];
                if (t == typeof(ulong)) { buffer[c] = Convert.ToUInt64(args[c]); }
                else if (t == typeof(uint)) { buffer[c] = Convert.ToUInt32(args[c]); }
                else if (t == typeof(ushort)) { buffer[c] = Convert.ToInt16(args[c]); }
                else if (t == typeof(long)) { buffer[c] = Convert.ToInt64(args[c]); }
                else if (t == typeof(int)) { buffer[c] = Convert.ToInt32(args[c]); }
                else if (t == typeof(short)) { buffer[c] = Convert.ToInt16(args[c]); }
                else if (t == typeof(sbyte)) { buffer[c] = Convert.ToSByte(args[c]); }
                else if (t == typeof(byte)) { buffer[c] = Convert.ToByte(args[c]); }
                else if (t == typeof(bool)) { buffer[c] = Convert.ToBoolean(args[c]); }
                else if (t == typeof(decimal)) { buffer[c] = Convert.ToDecimal(args[c]); }
                else if (t == typeof(double)) { buffer[c] = Convert.ToDouble(args[c]); }
                else if (t == typeof(float)) { buffer[c] = Convert.ToSingle(args[c]); }
                else if (t == typeof(IntPtr)) { buffer[c] = (IntPtr)Convert.ToInt64(args[c]); }
                else { buffer[c] = args[c]; }
            }

            //Return the buffer
            return buffer;
        }

        private long addClrObject(object instance) { 
            //Get the maximum ID
            long maxId = 0;
            foreach (clrObj o in p_ClrObjects) {
                if (o.ID >= maxId) { maxId = o.ID + 1; }
            }

            //Add the entry
            Array.Resize(ref p_ClrObjects, p_ClrObjects.Length + 1);
            p_ClrObjects[p_ClrObjects.Length - 1] = new clrObj {
                ID = maxId,
                Instance = instance
            };

            //Return the id
            return maxId;
        }
        private object getClrObject(long cid) { 
            //Look for the object
            foreach (clrObj o in p_ClrObjects) {
                if (o.ID == cid) { return o.Instance; }
            }

            //Not found
            throw new Exception("Not found");
        }
        private Type clrGetType(Assembly assembly, string type) {
            //Attempt to return the type
            if (assembly.GetType(type) != null) { return assembly.GetType(type); }

            //Look for the namespace shortcuts
            foreach (string s in p_ClrImports) {
                if (assembly.GetType(s + "." + type) != null) {
                    return assembly.GetType(s + "." + type);
                }
            }

            //Not found
            throw new Exception("Type not found");
        }

        public void clr_remove(long cid) { 
            //Define the new array buffer
            clrObj[] buffer = new clrObj[p_ClrObjects.Length - 1];
            int ai = 0;
            bool found = false;

            //Remove the entry
            for (int c = 0; c < buffer.Length; c++) {
                if (p_ClrObjects[c].ID == cid) { found = true; continue; }
                buffer[ai] = p_ClrObjects[c];
                ai++;
            }
            
            //Found?
            if (!found) {
                throw new Exception("CLR Object was not found");
            }

            //Set the array to the buffer
            p_ClrObjects = buffer;
        }
        public void clr_useNamespace(string name) { 
            //Already in use?
            foreach (string s in p_ClrImports) {
                if (s == name) { throw new Exception("Namespace already in use"); }
            }

            //Add the name to the clr imports
            Array.Resize(ref p_ClrImports, p_ClrImports.Length + 1);
            p_ClrImports[p_ClrImports.Length - 1] = name;
        }
        public long clr_addNetAssembly(string name) { 
            //Define the directory of the .net libraries
            string netPath = new FileInfo(Environment.SystemDirectory + "../../Microsoft.NET/Framework/v2.0.50727/" + name + ".dll").FullName;

            //Create the assembly
            return addClrObject(Assembly.LoadFile(netPath));
        }
        public long clr_addAssembly(string filename) { 
            //Create the assembly
            return addClrObject(Assembly.LoadFile(new FileInfo(filename).FullName));
        }
        public long clr_createInstance(string asmId, string type, params object[] args) { 
            //Get the assembly
            long assembly = Convert.ToInt64(p_State.Script.getValueFromRaw(asmId, p_State));
            type = p_State.Script.getValueFromRaw(type, p_State).ToString();
            Assembly asm = (Assembly)getClrObject(assembly);
            Type t = clrGetType(asm, type);

            //Create the instance
            object instance = null;
            foreach (ConstructorInfo ci in t.GetConstructors()) {
                ParameterInfo[] p = ci.GetParameters();
                if (p.Length != args.Length) { continue; }
                try {
                    instance = Activator.CreateInstance(t, clrConvMethodTypes(args, ci.GetParameters()));
                    break;
                }
                catch { }
            }

            //Push the instance to the current definition
            Interpreter.definition def = new Interpreter.definition { 
                ValueInstance = instance,
                Name = p_State.CurrentDefinition.Name
            };
            def.setDef(DefinitionType.CommonLanguageRuntimeObject);
            p_State.CurrentDefinition = def;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long clr_invoke(string asmId, string type, string name, params object[] args) {
            //Grab the static class
            type = p_State.Script.getValueFromRaw(type, p_State).ToString();
            name = p_State.Script.getValueFromRaw(name, p_State).ToString();
            long assembly = Convert.ToInt64(p_State.Script.getValueFromRaw(asmId, p_State));
            Assembly asm = (Assembly)getClrObject(assembly);
            Type t = clrGetType(asm, type);

            //Invoke the function
            foreach (MethodInfo m in t.GetMethods()) {
                if (m.Name == name) {
                    try {
                        object ret = m.Invoke(null, clrConvMethodTypes(args, m.GetParameters()));
                        
                        //Apply the result to the definition
                        p_State.CurrentDefinition = new Interpreter.definition() { 
                            Name = p_State.CurrentDefinition.Name,
                            ValueInstance = ret                           
                        };
                        ((Interpreter.definition)p_State.CurrentDefinition).setDef(DefinitionType.CommonLanguageRuntimeObject);
                        p_State.ApplyDefinitionGlobal = true;
                        return 0;
                    }
                    catch { }
                }
            }
            
            //Method not found
            throw new Exception("Method was not found");
        }
        public long clr_invoke(string vName, string method, params object[] args) { 
            //Get the object
            method = p_State.Script.getValueFromRaw(method, p_State).ToString();
            object instance = p_State.Script.getValueFromRaw(vName, p_State);
            
            //Invoke the method
            foreach (MethodInfo m in instance.GetType().GetMethods()) {
                if (m.Name == method) {
                    try {
                        object ret = m.Invoke(instance, clrConvMethodTypes(args, m.GetParameters()));

                        //Apply the result to the definition
                        p_State.CurrentDefinition = new Interpreter.definition() {
                            Name = p_State.CurrentDefinition.Name,
                            ValueInstance = ret
                        };
                        ((Interpreter.definition)p_State.CurrentDefinition).setDef(DefinitionType.CommonLanguageRuntimeObject);
                        p_State.ApplyDefinitionGlobal = true;
                        return 0;
                    }
                    catch { }
                }
            }

            //Not found
            throw new Exception("Method was not found");
        }
        public long clr_getValue(string vName, string variable) { 
            //Get the instance
            variable = p_State.Script.getValueFromRaw(variable, p_State).ToString();
            object instance = p_State.Script.getValueFromRaw(vName, p_State);

            //Define the recieved value
            object value = null;
            bool found = false;

            //Get the value
            foreach (PropertyInfo p in instance.GetType().GetProperties()) {
                if (p.Name == variable) {
                    value = p.GetValue(instance, null);
                    found = true;
                    break;
                }
            }
            if (!found) { 
                foreach (FieldInfo f in instance.GetType().GetFields()) {
                    if (f.Name == variable) {
                        value = f.GetValue(instance);
                        found = true;
                        break;
                    }
                } 
            }

            //Found?
            if (!found) {
                throw new Exception("Property was not found");
            }

            //Set the definition value
            Interpreter.definition def = new Interpreter.definition() { 
                Name = p_State.CurrentDefinition.Name,
                ValueInstance = value
            };
            def.setDef(DefinitionType.CommonLanguageRuntimeObject);
            p_State.CurrentDefinition = def;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long clr_getValue(string asmId, string type, string name) { 
            //Get the values
            asmId = p_State.Script.getValueFromRaw(asmId, p_State).ToString();
            type = p_State.Script.getValueFromRaw(type, p_State).ToString();
            name = p_State.Script.getValueFromRaw(name, p_State).ToString();
            
            //Get the type
            Type t = clrGetType((Assembly)getClrObject(Convert.ToInt64(asmId)), type);
            
            //Get the value
            object value = null;
            bool found = true;
            foreach (PropertyInfo p in t.GetProperties()) {
                if (p.Name == name) {
                    value = p.GetValue(null, null);
                    found = true;
                    break;
                }
            }
            if (!found) {
                foreach (FieldInfo f in t.GetFields()) {
                    if (f.Name == name) {
                        value = f.GetValue(null);
                        found = true;
                        break;
                    }
                }
            }

            //Found?
            if (!found) {
                throw new Exception("Property was not found");
            }

            //Push the value to the current definition
            Interpreter.definition def = new Interpreter.definition() {
                Name = p_State.CurrentDefinition.Name,
                ValueInstance = value,
            };
            def.setDef(DefinitionType.CommonLanguageRuntimeObject);
            p_State.CurrentDefinition = def;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public void clr_setValue(string vName, string variable, string valueStr) { 
            //Get the values
            object value = p_State.Script.getValueFromRaw(valueStr, p_State);
            variable = p_State.Script.getValueFromRaw(variable, p_State).ToString();
            object instance = p_State.Script.getValueFromRaw(vName, p_State);

            //Set the value
            foreach (PropertyInfo p in instance.GetType().GetProperties()) {
                if (p.Name == variable) {
                    p.SetValue(instance, value, null);
                    return;
                }
            }
            foreach (FieldInfo f in instance.GetType().GetFields()) {
                if (f.Name == variable) {
                    f.SetValue(instance, value);
                    return;
                }
            }

            //Not found
            throw new Exception("Value was not found");
        }
        public void clr_setValue(string asmId, string type, string variable, string valueStr) { 
            //Get the values
            object value = p_State.Script.getValueFromRaw(valueStr, p_State);
            long id = Convert.ToInt64(p_State.Script.getValueFromRaw(asmId, p_State));
            type = p_State.Script.getValueFromRaw(type, p_State).ToString();
            variable = p_State.Script.getValueFromRaw(variable, p_State).ToString();

            //Get the type
            Type t = clrGetType((Assembly)getClrObject(id), type);

            //Set the value
            foreach (PropertyInfo p in t.GetProperties()) {
                if (p.Name == variable) {
                    p.SetValue(null, value, null);
                    return;
                }
            }
            foreach (FieldInfo f in t.GetFields()) {
                if (f.Name == variable) {
                    f.SetValue(null, value);
                    return;
                }
            }

            //Not found
            throw new Exception("Value not found");
        }
        public long clr_createHandle(string vName) { 
            //Create a reference to the instance
            return addClrObject(p_State.Script.getValueFromRaw(vName, p_State));
        }
        public long clr_fromHandle(long handle) { 
            //Get the object
            object inst = getClrObject(handle);

            //Set the current definition to the CLR object instance
            p_State.CurrentDefinition = new Interpreter.definition { 
                Name = p_State.CurrentDefinition.Name,
                ValueInstance = inst
            };
            ((Interpreter.definition)p_State.CurrentDefinition).setDef(DefinitionType.CommonLanguageRuntimeObject);
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long clr_listMethods(long assembly, string type) { 
            //Get the type
            Assembly asm = (Assembly)getClrObject(assembly);
            Type t = clrGetType(asm, type);

            //Define the array to apply to the definition
            object[] array = new object[0];
            foreach (MethodInfo m in t.GetMethods()) {
                if (GenericMethods.ignoreMethod(m)) { continue; }
                bool exist = false;
                foreach (object s in array) {
                    if (s.ToString() == m.Name) { exist = true; break; }
                }
                if (exist) { continue; }
                Array.Resize(ref array, array.Length + 1);
                array[array.Length - 1] = m.Name;
            }

            //Push the array to the definition
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long clr_listProperties(long assembly, string type) { 
            //Get the type
            Assembly asm = (Assembly)getClrObject(assembly);
            Type t = clrGetType(asm, type);

            //Define the array to store the properties
            object[] array = new object[0];

            //Loop through the properties
            foreach (PropertyInfo prop in t.GetProperties()) { 
                //Add
                Array.Resize(ref array, array.Length + 1);
                array[array.Length - 1] = prop.Name;
            }

            //Push the array to the definition
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long clr_listFields(long assembly, string type) { 
            //Get the type
            Assembly asm = (Assembly)getClrObject(assembly);
            Type t = clrGetType(asm, type);

            //Define the array to push to the defintion
            object[] array = new object[0];

            //Fill in the array with tehe fields
            foreach (FieldInfo f in t.GetFields()) {
                Array.Resize(ref array, array.Length + 1);
                array[array.Length - 1] = f.Name;
            }

            //Push the array to the current definition
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public string clr_toString(string name) {
            //Return the CLR ToString result
            return p_State.Script.getValueFromRaw(name, p_State).ToString();
        }
        
        private class clrObj {
            public long ID;
            public object Instance;
        }
    }
}