﻿using System;
using System.IO;
using System.Collections;

namespace Plywood {
    public partial class GenericMethods {
        public long fs_physical_createdTime(string filename) { return getFSEntryInfo(filename).CreationTime.Ticks; }
        public long fs_physical_accessedTime(string filename) { return getFSEntryInfo(filename).LastAccessTime.Ticks; }
        public long fs_physical_modifiedTime(string filename) { return getFSEntryInfo(filename).LastWriteTime.Ticks; }
        public void fs_physical_delete(string filename) {
            //Delete directory?
            if (!File.Exists(filename)) {
                //Delete the directory
                foreach (DirectoryInfo d in new DirectoryInfo(filename).GetDirectories()) {
                    fs_physical_delete(d.FullName);
                }
                foreach (FileInfo f in new DirectoryInfo(filename).GetFiles()) {
                    File.Delete(f.FullName);
                }
                Directory.Delete(filename);
                return;
            }

            //Delete the file
            File.Delete(filename);
        }
        public void fs_physical_createDir(string dir) {
            //Create the directory
            Directory.CreateDirectory(dir);
        }
        public long fs_physical_size(string filename) {
            //Get the file size
            return new FileInfo(filename).Length;
        }
        public void fs_physical_move(string from, string to) {
            //Is it a file?
            if (File.Exists(from)) { File.Move(from, to); return; }

            //Move the directory
            fs_dirEntryCopyCallback(from, from, to, true, delegate(string f, string t, bool isDir, string hostPath) {
                //Is it a directory?
                if (isDir) {
                    //Just create it
                    try { Directory.CreateDirectory(t); }
                    catch { }
                    return;
                }

                //Move the file
                File.Move(f, t);
            });

            //Delete the folder
            fs_physical_delete(from);
        }
        public void fs_physical_copy(string from, string to) {
            //Is it a file?
            if (File.Exists(from)) { File.Copy(from, to); return; }

            //Copy the directory
            fs_dirEntryCopyCallback(from, from, to, true, delegate(string f, string t, bool isDir, string host) {
                //Is it a directory?
                if (isDir) {
                    //Create it
                    try { Directory.CreateDirectory(t); }
                    catch { }
                    return;
                }

                //Copy the file
                File.Copy(f, t);
            });
        }
        public string fs_physical_currentDirectory() {
            //Return the current directory
            return Environment.CurrentDirectory;
        }
        public void fs_physical_currentDirectory(string directory) {
            //Set the current directory
            Environment.CurrentDirectory = directory;
        }
        public string fs_physical_read(string filename) {
            //Read the file data as a string
            return File.ReadAllText(filename);
        }
        public void fs_physical_write(string filename, string contents) {
            //Write to the file
            File.WriteAllText(filename, contents);
        }
        public string fs_physical_getPath(long code) {
            return Environment.GetFolderPath((Environment.SpecialFolder)code);
        }
        public bool fs_physical_fileExists(string filename) { return File.Exists(filename); }
        public bool fs_physical_directoryExists(string path) { return Directory.Exists(path); }
        public long fs_physical_listFiles(string path) {
            //List the file list into an array
            FileInfo[] list = new DirectoryInfo(path).GetFiles();

            //Convert this to an object array
            object[] array = new object[list.Length];
            for (int c = 0; c < array.Length; c++) {
                array[c] = list[c].FullName;
            }

            //Push the array to the current defined value
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.ApplyDefinitionGlobal = true;
            p_State.CurrentDefinition.IsNumeric = false;
            return 0;
        }
        public long fs_physical_listDirectories(string path) {
            //List the directory list into a directory array
            DirectoryInfo[] list = new DirectoryInfo(path).GetDirectories();

            //Convert this to an object array
            object[] array = new object[list.Length];
            for (int c = 0; c < array.Length; c++) {
                array[c] = list[c].FullName;
            }

            //Push this array into the current definition
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public long fs_physical_listDrives() { 
            //Define the array to push to the definition
            DriveInfo[] drives = DriveInfo.GetDrives();
            object[] buffer = new object[drives.Length];
            for (int c = 0; c < buffer.Length; c++) {
                buffer[c] = drives[c].RootDirectory.FullName;
            }

            //Push the array to the current definition
            p_State.CurrentDefinition.ArrayList = (IList)buffer;
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public string fs_physical_getDriveVolumeLabel(string path) { return new DriveInfo(path).VolumeLabel; }
        public long fs_physical_getDriveCapacity(string path) { return new DriveInfo(path).TotalSize; }
        public long fs_physical_getDriveAvailable(string path) { return new DriveInfo(path).AvailableFreeSpace; }
        public string fs_physical_getDriveType(string path) { return new DriveInfo(path).DriveType.ToString(); }
        public string fs_physical_getDriveFormat(string path) { return new DriveInfo(path).DriveFormat.ToString(); }
        public string fs_physical_fileExtension(string path) { return new FileInfo(path).Extension; }
        public string fs_physical_getPathFullName(string path) { 
            //Get the full path
            if (File.Exists(path)) { return new FileInfo(path).FullName; }
            return new DirectoryInfo(path).FullName;
        }
        public string fs_physical_getSystemDirectory() { return Environment.SystemDirectory; }
        public long fs_physical_createStream(string filename, string mode) { 
            //Create a file stream
            return ioAdd(new FileStream(filename, (FileMode)Enum.Parse(typeof(FileMode), mode)));
        }

        private void fs_dirEntryCopyCallback(string hostPath, string from, string to, bool includeSubDir, dirEntryCopyCallback callback) {
            //Do we have access to the directory?
            try { new DirectoryInfo(from).GetDirectories(); }
            catch { return; }

            //Define the full physical paths
            from = new DirectoryInfo(from).FullName;
            to = new DirectoryInfo(to).FullName;
            hostPath = new DirectoryInfo(hostPath).FullName;

            //Process sub directories
            if (includeSubDir) {
                try { new DirectoryInfo(from).GetDirectories(); }
                catch { return; }
                foreach (DirectoryInfo d in new DirectoryInfo(from).GetDirectories()) {
                    fs_dirEntryCopyCallback(hostPath, d.FullName, d.FullName.Replace(from, to + "\\"), includeSubDir, callback);
                }
            }

            //Create this directory
            callback(from, to, true, hostPath);

            //Check access to the file list
            try { new DirectoryInfo(from).GetFiles(); }
            catch { return; }

            //Iterate over the files
            foreach (FileInfo f in new DirectoryInfo(from).GetFiles()) {
                callback(f.FullName, f.FullName.Replace(from, to + "\\"), false, hostPath);
            }
        }
        private delegate void dirEntryCopyCallback(string fileFrom, string fileTo, bool isDirectory, string hostPath);
        private FileSystemInfo getFSEntryInfo(string path) {
            //Return the appropriate instance
            return (File.Exists(path) ? (FileSystemInfo)new FileInfo(path) : (FileSystemInfo)new DirectoryInfo(path));
        }            
    }
}