﻿using System;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Drawing2D;

namespace Plywood {
    public partial class GenericMethods {
        private static gdiObject[] p_GDIObjects = new gdiObject[0];

        #region Image
        public void gdi_saveImage(long bitmap, string filename) { 
            //Save it
            ((Bitmap)getGDIObjectFromID(bitmap).Instance).Save(filename);
        }
        public long gdi_loadImage(string filename) {
            //Create an image
           return addGDIObject((Bitmap)Image.FromFile(filename), gdiObjType.IMAGE);
        }
        public long gdi_imageSize(long gid) { 
            //Return the size
            Bitmap img = (Bitmap)getGDIObjectFromID(gid).Instance;
            return gdi_createSize2((long)img.Width, (long)img.Height);
        }
        public void gdi_disposeImage(long gid) { 
            //Just dispose of the image
            ((Bitmap)getGDIObjectFromID(gid).Instance).Dispose();
        }
        public long gdi_createImage(long width, long height) { 
            //Create a new bitmap
            return addGDIObject(new Bitmap((int)width, (int)height), gdiObjType.IMAGE);
        }
        public long gdi_createHBitmap(long handle) { 
            //Return the bitmap
            return addGDIObject(Image.FromHbitmap((IntPtr)handle), gdiObjType.IMAGE);
        }
        public long gdi_createHIcon(long handle) { 
            //Return the bitmap
            return addGDIObject(Bitmap.FromHicon((IntPtr)handle), gdiObjType.IMAGE);
        }
        public long gdi_imageHBitmap(long gid) {
            return ((Bitmap)getGDIObjectFromID(gid).Instance).GetHbitmap().ToInt64();
        }
        public long gdi_imageHIcon(long gid) {
            return ((Bitmap)getGDIObjectFromID(gid).Instance).GetHicon().ToInt64();
        }
        public long gdi_imageGetPixel(long gid, long position) { 
            //Grab the position
            Point pos = (Point)getGDIObjectFromID(gid).Instance;

            //Get the pixel
            return gdi_imageGetPixel(gid, (long)pos.X, (long)pos.Y);
        }
        public long gdi_imageGetPixel(long gid, long x, long y) { 
            //Get the color
            Color c = ((Bitmap)getGDIObjectFromID(gid).Instance).GetPixel((int)x, (int)y);

            //Return the color
            return gdi_createColor((long)c.A, (long)c.R, (long)c.G, (long)c.B);
        }
        public void gdi_imageSetPixel(long gid, long color, long position) { 
            //Get the position
            Point pos = (Point)getGDIObjectFromID(position).Instance;

            //Set the pixel
            gdi_imageSetPixel(gid, color, (long)pos.X, (long)pos.Y);
        }
        public void gdi_imageSetPixel(long gid, long color, long x, long y) { 
            //Set the pixel
            ((Bitmap)getGDIObjectFromID(gid).Instance).SetPixel(
                    (int)x, (int)y, (Color)getGDIObjectFromID(color).Instance);
        }
        public long gdi_imageGetSize(long gid) { 
            //Return the image size
            Bitmap bitmap = (Bitmap)getGDIObjectFromID(gid).Instance;
            return gdi_createSize2((long)bitmap.Width, (long)bitmap.Height);
        }
        public long gdi_imageGetWidth(long gid) { 
            //Return the width
            return (long)((Bitmap)getGDIObjectFromID(gid).Instance).Width;
        }
        public long gdi_imageGetHeight(long gid) { 
            //Return the height
            return (long)((Bitmap)getGDIObjectFromID(gid).Instance).Height;
        }
        #endregion
        
        #region Size2
        public long gdi_createSize2(long width, long height) { 
            //Create the size
            return addGDIObject(new Size((int)width, (int)height), gdiObjType.SIZE2);
        }
        public long gdi_size2Width(long sid) { 
            //Return the width value
            return (long)((Size)getGDIObjectFromID(sid).Instance).Width;
        }
        public long gdi_size2Height(long sid) { 
            //Return the height value
            return (long)((Size)getGDIObjectFromID(sid).Instance).Height;
        }
        #endregion
        
        #region Vector2
        public long gdi_createVector2(long x, long y) { 
            //Create the vector
            return addGDIObject(new Point((int)x, (int)y), gdiObjType.VECTOR2);
        }
        public long gdi_vector2X(long gid) { 
            //Return the x
            return (long)((Point)getGDIObjectFromID(gid).Instance).X;
        }
        public long gdi_vector2Y(long gid) { 
            //Get the y value
            return (long)((Point)getGDIObjectFromID(gid).Instance).Y;
        }
        #endregion

        #region Color
        public long gdi_createColor(long a, long r, long g, long b) { 
            //Create the color
            return addGDIObject(Color.FromArgb((int)a, (int)r, (int)g, (int)b), gdiObjType.COLOR);
        }
        public long gdi_colorA(long gid) { 
            //Get the alpha value
            return (long)((Color)getGDIObjectFromID(gid).Instance).A;
        }
        public long gdi_colorR(long gid) { 
            //Get the red value
            return (long)((Color)getGDIObjectFromID(gid).Instance).R;
        }
        public long gdi_colorG(long gid) { 
            //Return the green value
            return (long)((Color)getGDIObjectFromID(gid).Instance).G;
        }
        public long gdi_colorB(long gid) { 
            //Return the blue value
            return (long)((Color)getGDIObjectFromID(gid).Instance).B;
        }
        #endregion

        #region Graphics
        public long gdi_createRenderer(long bitmap) { 
            //Create graphic from the bitmap
            Bitmap b = (Bitmap)getGDIObjectFromID(bitmap).Instance;
            return addGDIObject(Graphics.FromImage(b), gdiObjType.RENDERER);
        }
        public long gdi_createRendererFromHwnd(long handle) { 
            //Create the graphics
            return addGDIObject(Graphics.FromHwnd((IntPtr)handle), gdiObjType.RENDERER);
        }
        public long gdi_createRenderFromHdc(long hdc) { 
            //Create the graphics
            return addGDIObject(Graphics.FromHdc((IntPtr)hdc), gdiObjType.RENDERER);
        }
        public void gdi_setRendererSmoothing(long gid, string mode) { 
            //Set the smoothing
            ((Graphics)getGDIObjectFromID(gid).Instance).SmoothingMode = 
                    (SmoothingMode)Enum.Parse(typeof(SmoothingMode), mode);
        }
        public void gdi_setRendererInterpolation(long gid, string mode) { 
            //Set the interpolation mode
            ((Graphics)getGDIObjectFromID(gid).Instance).InterpolationMode =
                    (InterpolationMode)Enum.Parse(typeof(InterpolationMode), mode);
        }
        public void gdi_setRendererTextRenderingHint(long gid, string mode) { 
            //Set the text rendering hint
            ((Graphics)getGDIObjectFromID(gid).Instance).TextRenderingHint =
                    (TextRenderingHint)Enum.Parse(
                        typeof(TextRenderingHint), mode);
        }
        public string gdi_getRendererSmoothing(long gid) { 
            //Get the smoothing
            return ((Graphics)getGDIObjectFromID(gid).Instance).SmoothingMode.ToString();
        }
        public string gdi_getRendererInterpolation(long gid) { 
            //Get the interpolation mode
            return ((Graphics)getGDIObjectFromID(gid).Instance).InterpolationMode.ToString();
        }
        public string gdi_getRendererTextRenderingHint(long gid) { 
            //Get the text rendering hint
            return ((Graphics)getGDIObjectFromID(gid).Instance).TextRenderingHint.ToString();
        }
        public void gdi_disposeRenderer(long gid) { 
            //Dipose of the renderer
            ((Graphics)getGDIObjectFromID(gid).Instance).Dispose();
        }
        public void gdi_drawImage(long gid, long img, long point, long size) { 
            //Draw the image
            ((Graphics)getGDIObjectFromID(gid).Instance).DrawImage(
                            (Image)getGDIObjectFromID(img).Instance,
                            new Rectangle(
                                    (Point)getGDIObjectFromID(point).Instance,
                                    (Size)getGDIObjectFromID(size).Instance));
        }
        public void gdi_fillRectangle(long renderer, long brush, long location, long size) {
            //Fill the rectangle
            ((Graphics)getGDIObjectFromID(renderer).Instance).FillRectangle(
                        (Brush)getGDIObjectFromID(brush).Instance,
                        new Rectangle(
                                (Point)getGDIObjectFromID(location).Instance,
                                (Size)getGDIObjectFromID(size).Instance));
        }
        public void gdi_fillEllipse(long renderer, long brush, long location, long size) { 
            //Render the ellipse
            ((Graphics)getGDIObjectFromID(renderer).Instance).FillEllipse(
                        (Brush)getGDIObjectFromID(brush).Instance,
                        new Rectangle(
                                (Point)getGDIObjectFromID(location).Instance,
                                (Size)getGDIObjectFromID(size).Instance));
        }
        public void gdi_drawRectangle(long renderer, long pen, long location, long size) { 
            //Draw the rectangle
            ((Graphics)getGDIObjectFromID(renderer).Instance).DrawRectangle(
                        (Pen)getGDIObjectFromID(pen).Instance,
                        new Rectangle(
                                (Point)getGDIObjectFromID(location).Instance,
                                (Size)getGDIObjectFromID(size).Instance));
        }
        public void gdi_drawEllipse(long renderer, long pen, long location, long size) { 
            //Draw the ellipse
            ((Graphics)getGDIObjectFromID(renderer).Instance).DrawEllipse(
                        (Pen)getGDIObjectFromID(pen).Instance,
                        new Rectangle(
                                (Point)getGDIObjectFromID(location).Instance,
                                (Size)getGDIObjectFromID(size).Instance));
        }
        public void gdi_clear(long renderer, long color) { 
            //Clear
            ((Graphics)getGDIObjectFromID(renderer).Instance).Clear((Color)getGDIObjectFromID(color).Instance);
        }
        #endregion

        #region Brush
        public long gdi_solidBrush(long color) { 
            //Create a solid brush
            return addGDIObject(new SolidBrush((Color)getGDIObjectFromID(color).Instance), gdiObjType.BRUSH);
        }
        public long gdi_linearGradient(long color1, long color2, long point1, long point2) { 
            //Create a linear gradient brush
            return addGDIObject(new LinearGradientBrush(
                    (Point)getGDIObjectFromID(point1).Instance,
                    (Point)getGDIObjectFromID(point2).Instance,
                    (Color)getGDIObjectFromID(color1).Instance,
                    (Color)getGDIObjectFromID(color2).Instance), gdiObjType.BRUSH);
        }
        #endregion

        #region Pen
        public long gdi_createPen(long brush) {
            return addGDIObject(new Pen(
                    ((Brush)getGDIObjectFromID(brush).Instance)), gdiObjType.PEN);
        }
        public double gdi_getPenWidth(long gid) { 
            //Get the pen width
            return (double)((Pen)getGDIObjectFromID(gid).Instance).Width;
        }
        public void gdi_setPenWidth(long gid, double width) { 
            //Set the pen width
            ((Pen)getGDIObjectFromID(gid).Instance).Width = (float)width;
        }
        public string gdi_getPenDashStyle(long gid) { 
            //Return the dash style
            return ((Pen)getGDIObjectFromID(gid).Instance).DashStyle.ToString();
        }
        public void gdi_setPenDashStyle(long gid, string style) { 
            //Set the dash style
            ((Pen)getGDIObjectFromID(gid).Instance).DashStyle = (DashStyle)Enum.Parse(typeof(DashStyle), style);
        }
        public string gdi_getPenDashCap(long gid) { 
            //Return the dash cap
            return ((Pen)getGDIObjectFromID(gid).Instance).DashCap.ToString();
        }
        public void gdi_setPenDashCap(long gid, string cap) { 
            //Set the dash cap
            ((Pen)getGDIObjectFromID(gid).Instance).DashCap = (DashCap)Enum.Parse(typeof(DashCap), cap);
        }
        public double gdi_getPenDashOffset(long gid) { 
            //Return the dash offset
            return (double)((Pen)getGDIObjectFromID(gid).Instance).DashOffset;
        }
        public void gdi_setPenDashOffset(long gid, double offset) { 
            //Set the dash offset
            ((Pen)getGDIObjectFromID(gid).Instance).DashOffset = (float)offset;
        }
        #endregion
              
        public void gdi_remove(long gid) { 
            //Hang?
            while (gdiObject.Hang) ;
            gdiObject.Hang = true;

            //Define the new buffer
            gdiObject[] buffer = new gdiObject[p_GDIObjects.Length - 1];
            int ai = 0;
            bool found = false;

            //Fill in the buffer but skip over the entry to remove
            for (int c = 0; c < p_GDIObjects.Length; c++) {
                if (p_GDIObjects[c].ID == gid) { found = true; continue; }
                buffer[ai] = p_GDIObjects[c];
                ai++;
            }
            if (!found) {
                throw new Exception("ID not found");
            }

            //Set the global array to the buffer
            p_GDIObjects = buffer;
            
            //Stop hang
            gdiObject.Hang = false;
        }

        public string gdi_getType(long gid) { 
            //Return the type
            return getGDIObjectFromID(gid).Type.ToString();
        }

        private long addGDIObject(object inst, gdiObjType type) { 
            //Hang?
            while (gdiObject.Hang) ;
            gdiObject.Hang = true;

            //Get the id
            long maxId = 0;
            foreach (gdiObject o in p_GDIObjects) {
                if (o.ID >= maxId) { maxId = o.ID + 1; }
            }

            //Add the instance
            Array.Resize(ref p_GDIObjects, p_GDIObjects.Length + 1);
            p_GDIObjects[p_GDIObjects.Length - 1] = new gdiObject() { 
                ID = maxId,
                Type = type, 
                Instance = inst
            };

            //Stop hang
            gdiObject.Hang = false;

            //Return the id
            return maxId;
        }
        private gdiObject getGDIObjectFromID(long gid) { 
            //Find the id
            foreach (gdiObject o in p_GDIObjects) {
                if (o.ID == gid) { return o; }
            }

            //Not found 
            throw new Exception("Not found");
        }

        private class gdiObject {
            public long ID;
            public object Instance;
            public gdiObjType Type;
            public static bool Hang;
        }

        private enum gdiObjType { 
            SIZE2 = 0,
            VECTOR2 = 1,
            COLOR = 2,
            IMAGE = 3,
            BRUSH = 4,
            RENDERER = 5,
            PEN = 6
        }
    }
}