﻿using System;
using System.IO;
using System.Collections;
using System.Net.Sockets;
using System.Text;

namespace Plywood {
    public partial class GenericMethods {
        private static socket[] p_Sockets = new socket[0];
        private static listener[] p_Listener = new listener[0];

        public long net_listenerCreate(string address, long port) { 
            //Define the listener
            TcpListener listen = new TcpListener(new System.Net.IPAddress(getIp(address)), (int)port);
          
            //Add the listener
            long maxId = 0;
            foreach (listener l in p_Listener) {
                if (l.ID >= maxId) { maxId = l.ID + 1; }
            }
            while (listener.Hang) ;
            listener.Hang = true;
            Array.Resize(ref p_Listener, p_Listener.Length + 1);
            p_Listener[p_Listener.Length - 1] = new listener() {
                ID = maxId,
                Listener = listen
            };
            listener.Hang = false;

            return maxId;
        }
        public void net_listenerStart(long lid) {
            //Start the listener
            getListenerFromId(lid).Listener.Start();
        }
        public long net_listenerAccept(long lid) { 
            //Create a new socket
            long sid = net_socketCreate();

            //Accept a socket
            getSockFromId(sid).Socket = getListenerFromId(lid).Listener.AcceptSocket();

            //Return the sid
            return sid;
        }
        public void net_listenerStop(long lid) {
            //Stop the listener
            getListenerFromId(lid).Listener.Stop();
        }
        public void net_listenerRemove(long lid) { 
            //Hang
            while (listener.Hang) ;
            listener.Hang = true;

            //Remove the listener
            listener[] buffer = new listener[p_Listener.Length - 1];
            int ai = 0;
            bool found = false;
            for (int c = 0; c < p_Listener.Length; c++) {
                if (p_Listener[c].ID == lid) { found = true; continue; }
                buffer[ai] = p_Listener[c];
                ai++;
            }
            if (!found) {
                throw new Exception("Listener was not found");
            }
            p_Listener = buffer;
            listener.Hang = false;
        }
        public string net_listenerLocalEndPoint(long lid) { 
            //Return the local end point
            return getListenerFromId(lid).Listener.LocalEndpoint.ToString();
        }

        public long net_createStream(long sid) { 
            //Get the socket
            Socket socket = getSockFromId(sid).Socket;

            //Create a stream
            return addNetSocket(socket);
        }

        private long addNetSocket(Socket socket) { 
            //Get the ID
            long maxID = 0;
            foreach (socket s in p_Sockets) {
                if (s.ID >= maxID) { maxID = s.ID + 1; }
            }

            //Add the socket
            Array.Resize(ref p_Sockets, p_Sockets.Length + 1);
            p_Sockets[p_Sockets.Length - 1] = new socket {
                ID = maxID,
                Socket = socket
            };

            //Return the ID
            return maxID;
        }

        public long net_socketCreate() { 
            //Hang?
            while (socket.Hang) ;
            socket.Hang = true;

            //Get the maximum socket id
            long maxId = 0;
            foreach (socket s in p_Sockets) {
                if (s.ID >= maxId) { maxId = s.ID + 1; }
            }

            //Create the sockets
            Array.Resize(ref p_Sockets, p_Sockets.Length + 1);
            p_Sockets[p_Sockets.Length - 1] = new socket {
                ID = maxId,
                Socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.IP)
            };

            //Stop hang
            socket.Hang = false;

            //Return the id of the socket
            return maxId;
        }
        public bool net_socketConnect(long sid, string host, long port) { 
            //Get the socket
            socket sock = getSockFromId(sid);

            //Connect
            try {
                sock.Socket.Connect(host, (int)port);
                return true;
            }
            catch { 
                //Failed to connect
                return false;
            }
        }
        public long net_socketSend(long sid, string message) { 
            //Get the sccket
            socket s = getSockFromId(sid);

            //Send the packet
            return (long)s.Socket.Send(Encoding.ASCII.GetBytes(message));
        }
        public long net_socketSendBytes(string sid, string definition) { 
            //Get the sid value
            long id = Convert.ToInt64(p_State.Script.getValueFromRaw(sid, p_State));

            //Get the definition
            IDefinition def = p_State.Script.getDefByName(definition);

            //Read the bytes
            object[] array = (object[])def.ArrayList;
            byte[] packet = new byte[array.Length];
            for (int c = 0; c < packet.Length; c++) {
                packet[c] = Convert.ToByte(array[c]);
            }

            //Send the bytes to the socket
            return (long)getSockFromId(id).Socket.Send(packet);
        }
        public string net_socketRecieve(long sid, long length)
        { 
            //Get the socket
            socket s = getSockFromId(sid);

            //Recieve the packet
            byte[] packet = new byte[length];
            int count = s.Socket.Receive(packet);

            //Resize the packet to it's recieved length
            Array.Resize(ref packet, count);
            
            //Return the packet
            return Encoding.ASCII.GetString(packet);
        }
        public long net_socketRecieveBytes(long sid, long length) {
            //Recieve the packet
            byte[] packet = new byte[length];
            getSockFromId(sid).Socket.Receive(packet);

            //Setup the definition
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.ApplyDefinitionGlobal = true;

            //Convert the packet to an object array
            object[] pObj = new object[packet.Length];
            for (int c = 0; c < packet.Length; c++) {
                pObj[c] = packet[c];
            }

            //Push the object array to the current definition
            p_State.CurrentDefinition.ArrayList = (IList)pObj;
            return 0;
        }
        public void net_socketDisconnect(long sid) { 
            //Disconnect from the socket
            getSockFromId(sid).Socket.Disconnect(true);
        }
        public void net_socketClose(long sid) { 
            //Close the socket
            getSockFromId(sid).Socket.Close();
        }
        public void net_socketRemove(long sid) {
            //Hang?
            while (socket.Hang) ;
            socket.Hang = true;

            //Define the buffer for the socket
            socket[] buffer = new socket[p_Sockets.Length - 1];
            int ai = 0;
            bool found = false;

            //Fill in the buffer
            for (int c = 0; c < p_Sockets.Length; c++) {
                if (p_Sockets[c].ID == sid) { found = true; continue; }
                buffer[ai] = p_Sockets[c];
                ai++;
            }

            //Was it removed?
            if (!found) {
                throw new Exception("Socket was not found");
            }

            //Push the buffer to the global
            p_Sockets = buffer;

            //Stop hang
            socket.Hang = false;
        }
        public string net_socketLocalEndPoint(long sid) {
            //Get the local end point 
            return getSockFromId(sid).Socket.LocalEndPoint.ToString();
        }
        public string net_socketRemoteEndPoint(long sid) { 
            //Get the remote end point
            return getSockFromId(sid).Socket.RemoteEndPoint.ToString();
        }

        private byte[] getIp(string ip) { 
            //Split it up
            string[] split = ip.Split('.');

            //Check 
            if (split.Length != 4) {
                throw new Exception("Invalid IP");
            }
             
            //Define the return buffer
            byte[] buffer = new byte[4];
            for (int c = 0; c < 4; c++) {
                buffer[c] = Convert.ToByte(split[c]);
            }

            //Return it
            return buffer;
        }

        private socket getSockFromId(long sid) { 
            //Look for the socket
            foreach (socket s in p_Sockets) {
                if (s.ID == sid) { return s; }
            }

            //Not found
            throw new Exception("Socket not found");
        }
        private listener getListenerFromId(long lid) { 
            //Look for the listener
            foreach (listener l in p_Listener) {
                if (l.ID == lid) { return l; }
            }

            //Not found
            throw new Exception("Listener was not found");
        }

        private class socket {
            public long ID;
            public Socket Socket;
            public static bool Hang;
        }
        private class listener {
            public long ID;
            public TcpListener Listener;
            public static bool Hang;
        }
        private class sockStream : Stream {
            private Socket p_Socket;
            public sockStream(Socket socket) { 
                //Define the globals
                p_Socket = socket;
            }

            public override void Flush() {
                throw new NotSupportedException();
            }
            public override void Write(byte[] buffer, int offset, int count) {
                //Write to the socket
                p_Socket.Send(buffer, offset, count, SocketFlags.None);
            }
            public override void WriteByte(byte value) {
                //Let the local mehod handle this
                Write(new byte[] { value }, 0, 1);
            }
            public override long Seek(long offset, SeekOrigin origin) {
                throw new NotSupportedException();
            }
            public override int Read(byte[] buffer, int offset, int count) {
                //Read 
                return p_Socket.Receive(buffer, offset, count, SocketFlags.None);
            }
            public override int ReadByte() {
                //Define the buffer to read from
                byte[] r = new byte[1];
                Read(r, 0, 1);

                //Return the first byte
                return (int)r[0];
            }
            public override void SetLength(long value)  {
                throw new NotSupportedException();
            }

            public override bool CanRead { get { return true; } }
            public override bool CanSeek { get { return false; } }
            public override bool CanWrite { get { return true; } }

            public override long Length { get { throw new NotSupportedException(); } }
            public override long Position { get { throw new NotSupportedException(); } set { throw new NotSupportedException(); } }
        }   
    }
}