﻿using System;
using System.IO;
using System.Text;
using System.Collections;

namespace Plywood {
    public partial class GenericMethods {
        private static ioObj[] p_Streams = new ioObj[0];

        public long io_create() { 
            //Create a memory stream
            return ioAdd(new MemoryStream());
        }
        public void io_close(long iid) { 
            //Close the stream
            getStream(iid).Close();
        }
        public void io_remove(long iid) { 
            //Remove it
            ioObj[] buffer = new ioObj[p_Streams.Length - 1];
            int ai = 0;
            bool found = false;
            for (int c = 0; c < p_Streams.Length; c++) {
                if (p_Streams[c].ID == iid) { found = true; continue; }
                buffer[ai] = p_Streams[c];
                ai++;
            }

            //Found?
            if (!found) {
                throw new Exception("Stream not found");
            }

            //Clean up
            p_Streams = buffer;
        }
        public void io_write(long iid, string data) { 
            //Write the data
            getStream(iid).Write(Encoding.ASCII.GetBytes(data), 0, data.Length);
        }
        public void io_writeByte(long iid, long value) { 
            //Write the byte
            getStream(iid).WriteByte((byte)value);
        }
        public void io_writeArray(string id, string definition, string offset, string length) { 
            //Get the definition
            IDefinition d = p_State.Script.getDefByName(definition);

            //Get the ID, length and offset
            long iid = Convert.ToInt64(p_State.Script.getValueFromRaw(id, p_State));
            long len = Convert.ToInt64(p_State.Script.getValueFromRaw(length, p_State));
            long off = Convert.ToInt64(p_State.Script.getValueFromRaw(offset, p_State));

            //Get the array
            object[] array = (object[])d.ArrayList;

            //Convert to bytes
            byte[] bytes = new byte[array.Length];
            for (int c = 0; c < array.Length; c++) {
                bytes[c] = Convert.ToByte(array[c]);
            }

            //Write the array
            getStream(iid).Write(bytes, (int)off, (int)len);
        }
        public long io_readByte(long iid) { 
            //Read the next byte
            return (long)getStream(iid).ReadByte();
        }
        public long io_read(long iid, long length) { 
            //Define the array to read to
            byte[] bytes = new byte[length];

            //Read them
            getStream(iid).Read(bytes, 0, bytes.Length);

            //Convert to a object array
            object[] array = new object[bytes.Length];
            for (int c = 0; c < bytes.Length; c++) {
                array[c] = (long)bytes[c];
            }

            //Push the array t the definition
            p_State.CurrentDefinition.Value = "ARRAY_DEF";
            p_State.CurrentDefinition.IsNumeric = false;
            p_State.CurrentDefinition.ArrayList = (IList)array;
            p_State.ApplyDefinitionGlobal = true;
            return 0;
        }
        public void io_seek(long iid, long offset, string seekPos) { 
            //Seek the stream
            getStream(iid).Seek(offset, (SeekOrigin)Enum.Parse(typeof(SeekOrigin), seekPos));
        }
        public long io_length(long iid) {
            //Return the stream length
            return getStream(iid).Length;
        }
        public void io_length(long iid, long length) { 
            //Set the stream length
            getStream(iid).SetLength(length);
        }

        private long ioAdd(Stream stream) { 
            //Find the ID
            long id = 0;
            foreach (ioObj io in p_Streams) {
                if (io.ID >= id) { id = io.ID + 1; }
            }

            //Add it
            Array.Resize(ref p_Streams, p_Streams.Length + 1);
            p_Streams[p_Streams.Length - 1] = new ioObj {
                ID = id,
                Stream = stream
            };
            return id;
        }
        private Stream getStream(long iid) { 
            //Find the ID
            foreach (ioObj o in p_Streams) {
                if (o.ID == iid) { return o.Stream; }
            }

            //Not found
            throw new Exception("Stream not found");
        }

        private class ioObj {
            public long ID;
            public Stream Stream;
        }
    }
}