﻿using System;
using System.IO;

namespace Plywood {
    public partial class GenericMethods {
        public void goLine(long line) {
            //Seek to the beginning
            p_State.MainStream.Seek(0, SeekOrigin.Begin);

            //Read to that line
            long lineBuffer = 0;
            while (lineBuffer != line && p_State.MainStream.Position < p_State.MainStream.Length) {
                p_State.Script.skipComment(p_State.MainStream);
                if (p_State.MainStream.ReadByte() == (byte)'\n') { lineBuffer++; }
            }
        }
        public void exit() {
            //Stop the execution
            p_State.Stop();
            p_State.ScreenBuffer.Close();
            File.Delete(p_State.ScreenBufferFilename);
        }
        public long lineCount() {
            //Buffer the stream position
            long pos = p_State.Stream.Position;

            //Seek to the beginning of the stream
            p_State.Stream.Seek(0, SeekOrigin.Begin);

            //Count ever new-line character
            long buffer = 0;
            while (p_State.Stream.Position < p_State.Stream.Length) {
                p_State.Script.skipComment(p_State.Stream);
                if (p_State.Stream.ReadByte() == (byte)'\n') { buffer++; }
            }

            //Push the position back
            p_State.Stream.Position = pos;

            //Return the buffer
            return buffer + 1;
        }
        public long currentLine() {
            //Buffer the stream position
            long strPos = p_State.Stream.Position;

            //Seek to the beginning of the stream
            p_State.Stream.Seek(0, SeekOrigin.Begin);

            //Get the number of newline character occurances
            long buffer = 0;
            while (p_State.Stream.Position != strPos) {
                if (p_State.Stream.ReadByte() == (byte)'\n') { buffer++; }
            }
            if (buffer != 0) { buffer--; }

            //Pop the stream position back
            p_State.Stream.Position = strPos;

            //Return the count
            return buffer;
        }
        public string fimport(string filename) {
            //Open the file
            FileStream fs = new FileStream(filename, FileMode.Open);

            //Create a new interpreter
            Interpreter i = new Interpreter();
          
            //Run the file
            string ret = i.Run(fs, true, new RunState(i, fs));

            //Add it's methods
            foreach (IMethod m in i.p_Methods) { 
                //Exist?
                bool exist = false;
                foreach (IMethod x in p_State.Script.p_Methods) {
                    if (x.Name == m.Name) { exist = true; break; }
                }
                if (exist) { continue; }
                p_State.Script.AddMethod(m);
            }

            //Return the interpretation result
            return ret;
        }
        public string execute(string script) {
            //Write the script to a stream
            Stream ms = p_State.Script.createStrStream(script);

            //Get a run state
            RunState state = getCloneState(ms);

            //Run it
            string ret = state.Script.Run(ms, false, state);

            //Apply definition changes
            state.Script.applyDefinitionChanges(p_State.Script, state.Script);
            return ret;
        }
        public string attempt(string script) {
            //Attempt to execute the script
            try {
                return execute(script);
            }
            catch { return ""; }
        }
        public void destruct(string name) {
            //Look for the definition
            int index = -1;
            for (int c = 0; c < p_State.Script.p_Defines.Length; c++) {
                if (p_State.Script.p_Defines[c].Name == name) {
                    index = c;
                    break;
                }
            }
            if (index == -1) {
                throw new Exception("Definition \"" + name + "\" not found");
            }

            //Rebuild the definitions
            IDefinition [] buffer = new IDefinition[p_State.Script.p_Defines.Length - 1];
            int ai = 0;
            for (int c = 0; c < p_State.Script.p_Defines.Length; c++) {
                if (c == index) { continue; }
                buffer[ai] = p_State.Script.p_Defines[c];
                ai++;
            }
            p_State.Script.p_Defines = buffer;
        }

        private RunState getCloneState(Stream stream) {
            //Define the run state
            Interpreter scr = (Interpreter)p_State.Script.Clone();
            RunState state = new RunState(scr, stream);
            state.MainStream = p_State.MainStream;
            return state;
        }

        public string @delegate() {
            //Read the scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Execute the scope
            return execute(scope);
        }
        public void @if(string logic) {
            //Logic true?
            if (!p_State.Script.getBoolMultiLogic(logic, p_State)) { return; }

            //Get the script inside the if scope
            string script = p_State.Script.skipScope(p_State.Stream);

            //Replace unwanted characters
            script = p_State.Script.remUnwantedCharacters(script);

            //Execute it
            execute(script);
        }
        public void nif(string logic) {
            //Logic false?
            if (p_State.Script.getBoolMultiLogic(logic, p_State)) { return; }

            //Read the scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Run the scope
            execute(scope);
        }
        public void @while(string logic) {
            //Read the whole scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);
            
            //Loop until the logic statement is false
            while (p_State.Script.getBoolMultiLogic(logic, p_State)) {              
                //Run the scope
                execute(scope);
            }
        }
        public void @for(string def, string logic, string incriment) {
            //Check the def code
            if (!def.StartsWith("def") || def.Split(' ').Length < 2) { throw new Exception("Invalid definition"); }

            //Grab the name of the defined value
            string variable = def.Split(' ')[1];

            //Execute the defitition code
            p_State.Script.Run(p_State.Script.createStrStream(def), false, p_State);

            //Grab the scope script
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Loop while logic expression is false
            while (p_State.Script.getBoolMultiLogic(logic, p_State)) {
                //Perform the scope script
                execute(scope);

                //Perform the incriment script
                p_State.Script.Run(p_State.Script.createStrStream(incriment), false, p_State);
            }

            //Destruct the defined variable
            destruct(variable);
        }
        public void @foreach(string def, string array) {
            //Get the definition to loop through
            IDefinition d = p_State.Script.getDefByName(array);

            //Check the definition code
            if (!def.StartsWith("def") || def.Split(' ').Length != 2) { throw new Exception("Invalid definition script"); }

            //Grab the defined value name
            string defName = def.Split(' ')[1];

            //Add the definition manually
            if (!p_State.Script.validName(defName)) {
                throw new Exception("Invalid definition value");
            }
            Array.Resize(ref p_State.Script.p_Defines, p_State.Script.p_Defines.Length + 1);
            p_State.Script.p_Defines[p_State.Script.p_Defines.Length - 1] = new Interpreter.definition() {
                Name = defName,
            };

            //Grab the scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Loop through the array
            for (int c = 0; c < d.ArrayList.Count; c++) {
                //Define the definition
                IDefinition define = p_State.Script.getDefByName(defName);

                //Grab the object from the array
                object o = d.ArrayList[c];

                //Set the numeric boolean flag
                define.IsNumeric = o.GetType() == typeof(long) || o.GetType() == typeof(double);

                //Set the value
                define.Value = o.ToString();

                //Push the value to the definition
                p_State.Script.applyDef(define);

                //Run the scope
                execute(scope);
            }

            //Remove the defined value
            destruct(defName);
        }
        public void @try() { @try(""); }
        public void @try(string catched) {
            //Read the scope
            string scope = p_State.Script.skipScope(p_State.Stream);
            scope = p_State.Script.remUnwantedCharacters(scope);

            //Attempt to perform the scope code
            try {
                Stream stream = p_State.Script.createStrStream(scope);
                RunState state = new RunState(p_State.Script, stream);
                state.MainStream = p_State.Stream;
                p_State.Script.Run(stream, false, state);
            }
            catch {
                //Perform the catch handler code
                execute(catched);
            }
        }            
    }
}