﻿using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Plywood {
    public partial class GenericMethods {
        public void sys_shell(string filename) { sys_shell(filename, ""); }
        public void sys_shell(string filename, string arguments) {
            //Define the process to run
            Process p = new Process();

            //Define if we should capture the process outputs
            bool capture = true;

            //Define the process startup settings
            p.StartInfo = new ProcessStartInfo {
                FileName = filename,
                Arguments = arguments,
                CreateNoWindow = capture,
                UseShellExecute = !capture,
                RedirectStandardOutput = capture,
                RedirectStandardError = capture
            };

            //Start the process
            p.Start();

            //Define the output streams?
            if (capture) {
                //Define the base delegate for printing
                DataReceivedEventHandler del = delegate(object sender, DataReceivedEventArgs e) {
                    //Ignore?
                    if (e.Data == null) { return; }

                    //Print it
                    print(e.Data + "\n");
                };

                //Set the output
                p.ErrorDataReceived += del;
                p.OutputDataReceived += del;

                //Begin listening
                p.BeginOutputReadLine();
            }

            //Wait for the application to exit
            p.WaitForExit();
        }
        public string sys_shellOut(string filename) { return sys_shellOut(filename, ""); }
        public string sys_shellOut(string filename, string arguments) { 
            //Define the process to run
            Process p = new Process();

            //Define the startup
            p.StartInfo = new ProcessStartInfo() { 
                FileName = filename, 
                Arguments = arguments,
                CreateNoWindow = true,
                UseShellExecute = false,
                RedirectStandardOutput = true,
                RedirectStandardError = true
            };

            //Define the output streams
            string buffer = "";
            DataReceivedEventHandler output = delegate(object sender, DataReceivedEventArgs e) {
                if (e.Data == null) { return; }
                buffer += e.Data + "\n";
            };
            p.OutputDataReceived += output;
            p.ErrorDataReceived += output;

            //Start it
            p.Start();
            p.BeginOutputReadLine();
            p.BeginErrorReadLine();
            p.WaitForExit();

            //Return the output
            if (buffer.Length != 0) {
                buffer = buffer.Substring(0, buffer.Length - 1);
            }
            return buffer;
        }

        
        public long sys_processGetId(string name) { return (long)Process.GetProcessesByName(name)[0].Id; }
        public string sys_processName(long id) { return Process.GetProcessById((int)id).ProcessName; }
        public void sys_processKill(long id) {
            //Kill the process
            Process.GetProcessById((int)id).Kill();
        }
        public long sys_processCurrent() {
            //Return the currrent proccess id
            return (long)Process.GetCurrentProcess().Id;
        }
        public long sys_processStart(long id) {
            //Get the time the process associated with the id was started
            return Process.GetProcessById((int)id).StartTime.Ticks;
        }
        public long sys_processMainWindowHandle(long id) { 
            //Return the process main window handle
            return Process.GetProcessById((int)id).MainWindowHandle.ToInt64();
        }
        public void sys_elevate() {
            //Let the local method handle this
            sys_elevate("");
        }
        public void sys_elevate(string noElevationAction) {
            //Already elevated?
            if (p_State.IsElevated) { return; }

            //Buffer the stream positon
            long strPos = p_State.Stream.Position;

            //Dump the defined variables and screen buffer to the variable buffer file
            FileStream varBuffer = new FileStream("varBuffer", FileMode.Create);
            string varComment = "/*\n" +
                                "     Script variable dump\n" +
                                "     Created: " + DateTime.Now.ToString() + "\n" +
                                "     Screen buffer size: " + p_State.ScreenBuffer.Length.ToString() + " bytes\n" +
                                "*/\n";
            varBuffer.Write(Encoding.ASCII.GetBytes(varComment), 0, varComment.Length);
            foreach (IDefinition d in p_State.Script.p_Defines) {
                if (d.IsDefaultGenerated) { continue; }
                string value = d.Value;
                if (d.IsNumeric) {
                    if (value.Length == 0) { value = "0"; }
                }
                else {
                    value = "\"" + value + "\"";
                }
                string line = "def " + d.Name + " = " + value + "\n";
                varBuffer.Write(Encoding.ASCII.GetBytes(line), 0, line.Length);
            }

            //Read the screen buffer to the file
            long screenBufferPos = p_State.ScreenBuffer.Position;
            p_State.ScreenBuffer.Seek(0, SeekOrigin.Begin);
            string printCommand = "print(\"";
            varBuffer.Write(Encoding.ASCII.GetBytes(printCommand), 0, printCommand.Length);
            while (p_State.ScreenBuffer.Position < p_State.ScreenBuffer.Length) {
                byte b = (byte)p_State.ScreenBuffer.ReadByte();
                bool last = (p_State.ScreenBuffer.Position == p_State.ScreenBuffer.Length - 1);
                if (b == (char)'\n') {
                    string newL = "\"&newl" + (!last ? "&\"" : "");
                    varBuffer.Write(Encoding.ASCII.GetBytes(newL), 0, newL.Length);
                    continue;
                }
                varBuffer.WriteByte(b);
            }
            printCommand = "\")\n";
            varBuffer.Write(Encoding.ASCII.GetBytes(printCommand), 0, printCommand.Length);
            varBuffer.Close();

            //Dump the stream to a file
            string file = "";
            Stream fs = FileSystem.CreateTempStream(out file);
            p_State.Stream.Seek(0, SeekOrigin.Begin);
            FileSystem.CopyStream(p_State.Stream, fs, 1024 * 1024);
            fs.Close();

            //Recreate this process
            Process p = new Process();
            p.StartInfo = new ProcessStartInfo() {
                FileName = Process.GetCurrentProcess().Modules[0].FileName,
                Arguments = "run " + file + " -d -p " + strPos.ToString(),
                Verb = "runas"
            };
            try { p.Start(); }
            catch {
                File.Delete("varBuffer");
                execute(noElevationAction);
                return;
            }

            //Kill this process
            try {
                p_State.ScreenBuffer.Close();
                File.Delete(p_State.ScreenBufferFilename);
            }
            catch { }
            Process.GetCurrentProcess().Kill();
        }
        public void sys_beep(long frequency, long duration) {
            //Cause beep
            Beep((int)frequency, (int)duration);
        }
        public bool sys_processExist(string name) {
            //Attempt to get the process
            try {
                return Process.GetProcessesByName(name).Length != 0;
            }
            catch { return false; }
        }

        [DllImport("kernel32.dll")]
        private static extern void Beep(int frequency, int duration);
    }
}