﻿using System;
using System.Text;
using System.IO;
using System.Threading;
using Plywood.Compiler;
using System.Drawing;
using System.Windows.Forms;

namespace Plywood {
    public partial class GenericMethods {
        public void printl(string line) { print(line + "\n"); }
        public void print(string line) { 
            //Print it
            p_State.doConsoleOp(TerminalMessageType.Write, line, false);
        }
        public string readl() { 
            //Read line
            return (string)p_State.doConsoleOp(TerminalMessageType.ReadLine, null, false);
        }
        public string read() { 
            //Read the key
            return p_State.doConsoleOp(TerminalMessageType.ReadKey, null, false).ToString();
        }

        public bool scr_cursorVisible() { return (bool)p_State.doConsoleOp(TerminalMessageType.CursorVisible, null, false); }
        public void scr_cursorVisible(bool val) { p_State.doConsoleOp(TerminalMessageType.CursorVisible, val, true); }
        public long scr_cursorX() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.CursorX, null, false); }
        public long scr_cursorY() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.CursorY, null, false); }
        public void scr_cursor(long x, long y) { 
            //Set the x and y
            p_State.doConsoleOp(TerminalMessageType.CursorX, (int)x, true);
            p_State.doConsoleOp(TerminalMessageType.CursorY, (int)y, true);
        }
        public void scr_cursorSize(long size) { p_State.doConsoleOp(TerminalMessageType.CursorSize, (int)size, true); }
        public long scr_cursorSize() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.CursorSize, null, false); }
        public long scr_bufferWidth() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.BufferWidth, null, false); }
        public long scr_bufferHeight() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.BufferHeight, null, false); }
        public void scr_bufferSize(long width, long height) { 
            //Set the buffer size
            p_State.doConsoleOp(TerminalMessageType.BufferWidth, (int)width, true);
            p_State.doConsoleOp(TerminalMessageType.BufferHeight, (int)height, true);
        }
        public long scr_windowWidth() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.WindowWidth, null, false); }
        public long scr_windowHeight() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.WindowHeight, null, false); }
        public void scr_windowSize(long width, long height) { 
            //Set the window size
            p_State.doConsoleOp(TerminalMessageType.WindowWidth, (int)width, true);
            p_State.doConsoleOp(TerminalMessageType.WindowHeight, (int)height, true);
        }
        public void scr_title(string title) { 
            //Set the title
            p_State.doConsoleOp(TerminalMessageType.Title, title, true);
        }
        public string scr_foreColor() { return p_State.doConsoleOp(TerminalMessageType.ForeColor, null, false).ToString(); }
        public string scr_backColor() { return p_State.doConsoleOp(TerminalMessageType.BackColor, null, false).ToString(); }
        public void scr_foreColor(string color) { 
            //Set the fore color
            p_State.doConsoleOp(TerminalMessageType.ForeColor, Enum.Parse(typeof(ConsoleColor), color), true);
        }
        public void scr_backColor(string color) { 
            //Set the back color
            p_State.doConsoleOp(TerminalMessageType.BackColor, Enum.Parse(typeof(ConsoleColor), color), true);
        }
        public long scr_maxWindowWidth() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.MaximumWindowWidth, null, false); }
        public long scr_maxWindowHeight() { return (long)(int)p_State.doConsoleOp(TerminalMessageType.MaximumWindowHeight, null, false); }
        public string scr_title() { 
            //Get the title
            return (string)p_State.doConsoleOp(TerminalMessageType.Title, null, false);
        }
        public void scr_clear() { p_State.doConsoleOp(TerminalMessageType.ClearScreen, null, false); }
        public long scr_createTerminalWindow() { return scr_createTerminalWindow(false); }
        public long scr_createTerminalWindow(bool exitWhenClose) { 
            //Create a terminal
            TerminalWindow wnd = null;
            bool added = false;

            //Create and show the form
            new Thread(new ThreadStart(delegate {
                wnd = new TerminalWindow();
                p_State.Script.AddTerminal(wnd);
                Form frm = wnd.CreateForm(exitWhenClose);
                frm.Load += delegate(object sender, EventArgs e) { added = true; };
                frm.ShowDialog();
            })).Start();
            while (!added) { Application.DoEvents(); Thread.Sleep(1000); }

            //Return the ID
            return p_State.Script.p_Terminals.Length - 1;
        }
        public void scr_terminalWindowBackColor(long index, long r, long g, long b) { 
            //Change the terminal back color
            ((TerminalWindow)p_State.Script.p_Terminals[index]).BackColor = Color.FromArgb(255, (int)r, (int)g, (int)b);
        }
        public void scr_terminalWindowForeColor(long index, long r, long g, long b) { 
            //Change the terminal fore color
            ((TerminalWindow)p_State.Script.p_Terminals[index]).ForeColor = Color.FromArgb(255, (int)r, (int)g, (int)b);
        }

        public void scr_terminalSelect(params object[] indexes) { 
            //Enable the appropriate terminals
            for (int c = 0; c < p_State.Script.p_Terminals.Length; c++) {
                bool enable = false;
                foreach (object o in indexes) { if (Convert.ToInt32(o) == c) { enable = true; } }
                p_State.Script.p_Terminals[c].Enabled = enable;                
            }
        }
        public void scr_primaryTerminal(long index) { RunState.PrimaryTerminal = (int)index; CommandSystem.Terminal = p_State.Script.p_Terminals[index]; }
        public long scr_primaryTerminal() { return (long)RunState.PrimaryTerminal; }
    }
}