﻿using System;
using System.IO;
using System.Windows.Forms;
using System.Reflection;
using System.Drawing;

namespace Plywood {
    public partial class GenericMethods {
        private static winComponent[] p_WindowComponents = new winComponent[0];

        public void win_msg(string message) {
            //Let the local method handle this
            win_msg(message, "Information");
        }
        public void win_msg(string message, string caption) {
            //Let the local method handle this
            win_msg(message, caption, (long)MessageBoxButtons.OK, (long)MessageBoxIcon.Information, "", "", "");
        }
        public void win_msg(string message, string caption, long type, long icon, string yes, string no, string cancel) {
            //Get the enum field from the types
            MessageBoxButtons buttons = (MessageBoxButtons)type;
            MessageBoxIcon icons = (MessageBoxIcon)icon;

            //Perform the message box
            DialogResult result = MessageBox.Show(message, caption, buttons, icons);

            //Was it yes/ok?
            if (result == DialogResult.Yes || result == DialogResult.OK) {
                //Perform the script for yes
                execute(yes);
                return;
            }

            //Was it no?
            if (result == DialogResult.Abort || result == DialogResult.No) {
                //Perform the script for no
                execute(no);
                return;
            }

            //Cancel?
            if (result == DialogResult.Cancel) {
                //Perform the script for cancel
                execute(cancel);
                return;
            }
        }
        public string win_openFile()  {
            //Let the local method handle this
            return win_openFile("Open file");
        }
        public string win_openFile(string title) {
            //Let the local method handle this
            return win_openFile(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments), title);
        }
        public string win_openFile(string startDir, string title) {
            //Let the local method handle this
            return win_openFile(startDir, title, "");
        }
        public string win_openFile(string startDir, string title, string filter) {
            //Let the local method handle this
            return win_openFile(startDir, title, filter, "");
        }
        public string win_openFile(string startDir, string title, string filter, string noFileSelected) {
            //Define the dialog
            OpenFileDialog o = new OpenFileDialog();

            //Define it's globals
            if (filter.Length != 0) { o.Filter = filter; }
            o.Title = title;
            o.InitialDirectory = new DirectoryInfo(startDir).FullName;

            //Open it
            if (o.ShowDialog() == DialogResult.OK) { return o.FileName; }

            //Perform the no file selected action
            execute(noFileSelected);
            return null;
        }
        public string win_saveFile() {
            //Let the local method handle this
            return win_saveFile("Save file");
        }
        public string win_saveFile(string title) {
            //Let the local method handle this
            return win_saveFile(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments), title);
        }
        public string win_saveFile(string startDir, string title) {
            //Let the local method handle this
            return win_saveFile(startDir, title, "");
        }
        public string win_saveFile(string startDir, string title, string filter) {
            //Let the local method handle this
            return win_saveFile(startDir, title, filter, "");
        }
        public string win_saveFile(string startDir, string title, string filter, string noFileSelected) {
            //Define the save dialog
            SaveFileDialog s = new SaveFileDialog();

            //Define its globals
            if (filter != "") { s.Filter = filter; }
            s.Title = title;
            s.InitialDirectory = new DirectoryInfo(startDir).FullName;

            //Show it
            if (s.ShowDialog() == DialogResult.OK) { return s.FileName; }

            //Run the no file selected script
            execute(noFileSelected);
            return null;
        }
        public string win_openDir() {
            //Let the local method handle this
            return win_openDir("");
        }
        public string win_openDir(string noDirSelected) {
            //Let the local method handle this
            return win_openDir(0, noDirSelected);
        }
        public string win_openDir(long startDir, string noDirSelected) {
            //Define the dialog
            FolderBrowserDialog f = new FolderBrowserDialog();

            //Define it's globals
            f.RootFolder = (Environment.SpecialFolder)startDir;

            //Show it
            if (f.ShowDialog() == DialogResult.OK) { return f.SelectedPath; }

            //Run the no directory selected script
            execute(noDirSelected);
            return null;
        }

        public long win_getComponentHandle(long wid) { 
            //Return the component handle
            return ((Control)winComponentFromId(wid).Instance).Handle.ToInt64();
        }
        public long win_getMousePosition() { 
            //Return the mouse position
            return gdi_createVector2(
                    (long)Cursor.Position.X,
                    (long)Cursor.Position.Y);
        }
        public string win_getMouseButton() {
            return Form.MouseButtons.ToString();
        }
        public long win_createComponent(string componentId) { 
            //Hang
            while (winComponent.Hang) ;
            winComponent.Hang = true;

            //Get the maximum ID
            long maxId = 0;
            foreach (winComponent w in p_WindowComponents) {
                if (w.ID >= maxId) { 
                    maxId = w.ID + 1; }
            }

            //Get the component type
            ComponentType cType = (ComponentType)Enum.Parse(typeof(ComponentType), componentId);

            //Get the object type
            Type oType = getTypeFromComponentType(cType);

            //Create the entry
            Array.Resize(ref p_WindowComponents, p_WindowComponents.Length + 1);
            p_WindowComponents[p_WindowComponents.Length - 1] = new winComponent() {
                ID = maxId,
                Instance = oType.Assembly.CreateInstance(oType.FullName),
                Type = cType
            };

            //Stop the hang
            winComponent.Hang = false;

            //Return the id
            return maxId;
        }
        public void win_deleteComponent(long wid) { 
            //Remove the id
            winComponent[] buffer = new winComponent[p_WindowComponents.Length - 1];
            int ai = 0;
            bool found = false;
            for (int c = 0; c < p_WindowComponents.Length; c++) {
                if (p_WindowComponents[c].ID == wid) { found = true; continue; }
                buffer[ai] = p_WindowComponents[c];
                ai++;
            }
            if (!found) {
                throw new Exception("Window component was not found");
            }
            p_WindowComponents = buffer;
        }
        public void win_addComponent(long parent, long wid) { 
            //Get the base control
            Control control = (Control)winComponentFromId(parent).Instance; 

            //Add the control
            control.Controls.Add((Control)winComponentFromId(wid).Instance);
            return;
        }
        public void win_removeComponent(long parent, long wid) { 
            //Remove the control
            ((Control)winComponentFromId(parent).Instance).Controls.Remove((Control)winComponentFromId(wid).Instance);
        }
        public void win_setComponentProperty(long wid, string name, string value) { 
            //Get the definition
            winComponent component = winComponentFromId(wid);

            //Set the property
            PropertyInfo prop = component.Instance.GetType().GetProperty(name);
            prop.SetValue(component.Instance, getPropValue(value, prop), null);
        }
        public void win_setComponentProperty(long wid, string name, long valueId) { 
            //Get the definition
            winComponent component = winComponentFromId(wid);

            //Set the property
            PropertyInfo prop = component.Instance.GetType().GetProperty(name);
            prop.SetValue(component.Instance, getGDIObjectFromID(valueId).Instance, null);
        }
        public object win_getComponentProperty(long wid, string name) { 
            //Get the property
            winComponent win = winComponentFromId(wid);
            PropertyInfo prop = win.Instance.GetType().GetProperty(name);

            //Get the object value
            object value = prop.GetValue(win.Instance, null);

            //Shall we create an id?
            if (value.GetType() == typeof(Color)) {
                Color clr = (Color)value;
                return gdi_createColor((long)clr.A, (long)clr.R, (long)clr.G, (long)clr.B);
            }
            else if (value.GetType() == typeof(Point)) {
                Point p = (Point)value;
                return gdi_createVector2((long)p.X, (long)p.Y);
            }
            else if (value.GetType() == typeof(Size)) {
                Size s = (Size)value;
                return gdi_createSize2((long)s.Width, (long)s.Height);
            }
            else if (value.GetType() == typeof(Bitmap)) {
                return addGDIObject(value, gdiObjType.IMAGE);
            }

            //Just return the value
            return value;
        }
        public void win_showForm(long wid) {
            //Show the form
            ((Form)winComponentFromId(wid).Instance).ShowDialog();
        }
        public void win_addComponentEvent(long wid, string name, string method) { 
            //Get the event 
            winComponent win = winComponentFromId(wid);
            EventInfo e = win.Instance.GetType().GetEvent(name);
                        
            //Get the handler type
            Type handler = e.EventHandlerType;

            //Setup the handler
            if (handler == typeof(EventHandler)) {
                e.AddEventHandler(win.Instance, new EventHandler(delegate(object se, EventArgs eA) {
                    execute(method + "()");
                }));
            }

        }
        public long win_addComponentFromHandle(long hwnd) { 
            //Create a new component for this handle
            while (gdiObject.Hang) ;
            winComponent.Hang = true;
            long maxId = 0;
            foreach (winComponent w in p_WindowComponents) {
                if (w.ID >= maxId) { maxId = w.ID + 1; }
            }
            Array.Resize(ref p_WindowComponents, p_WindowComponents.Length + 1);
            p_WindowComponents[p_WindowComponents.Length - 1] = new winComponent() { 
                ID = maxId,
                Instance = Control.FromHandle((IntPtr)hwnd),
                Type = ComponentType.Control
            };
            winComponent.Hang = false;
            return maxId;
        }
        public long win_clr(long wid) { 
            //Get the instance
            object instance = winComponentFromId(wid).Instance;

            //Push the instance to the current definition
            p_State.CurrentDefinition = new Interpreter.definition { 
                Name = p_State.CurrentDefinition.Name,
                ValueInstance = instance
            };
            ((Interpreter.definition)p_State.CurrentDefinition).setDef(DefinitionType.CommonLanguageRuntimeObject);
            return 0;            
        }

        private Type getTypeFromComponentType(ComponentType t) { 
            //Return the appropriate type
            switch (t) {
                case ComponentType.Form: return typeof(Form);
                case ComponentType.Panel: return typeof(Panel);
                case ComponentType.Control: return typeof(Control);
                case ComponentType.PictureBox: return typeof(PictureBox);
                case ComponentType.CheckBox: return typeof(CheckBox);
                case ComponentType.RadioBox: return typeof(RadioButton);
                case ComponentType.Button: return typeof(Button);
                case ComponentType.Label: return typeof(Label);
                case ComponentType.HScrollBar: return typeof(HScrollBar);
                case ComponentType.VScrollBar: return typeof(VScrollBar);
                case ComponentType.TextBox: return typeof(TextBox);
                case ComponentType.NumericUpDown: return typeof(NumericUpDown);
                case ComponentType.ProgressBar: return typeof(ProgressBar);
                case ComponentType.MenuStrip: return typeof(MenuStrip);
                case ComponentType.MenuStripButton: return typeof(ToolStripButton);
                case ComponentType.MenuStripLabel: return typeof(ToolStripLabel);
                case ComponentType.MenuStripIcon: return typeof(ToolStripItem);
                case ComponentType.MenuStripDropDownItem: return typeof(ToolStripDropDown);
                case ComponentType.StatusStrip: return typeof(StatusBar);
                case ComponentType.StatusStripProgressBar: return typeof(ToolStripProgressBar);
                case ComponentType.ToolStripMenu: return typeof(ToolStrip);
                case ComponentType.RichTextBox: return typeof(RichTextBox);
                case ComponentType.ContextMenuStrip: return typeof(ContextMenu);
                default:
                    throw new Exception("Type not supported");
            }
        }

        private winComponent winComponentFromId(long wid) { 
            //Find the id
            foreach (winComponent win in p_WindowComponents) {
                if (win.ID == wid) { return win; }
            }

            //Not found
            throw new Exception("Not found");
        }
        
        private class winComponent {
            public object Instance;
            public ComponentType Type;
            public long ID;
            public static bool Hang;
        }

        private enum ComponentType {
            Form = 0,
            Panel = 1,
            Control = 2,
            PictureBox = 3,
            CheckBox = 4,
            RadioBox = 5,
            Button = 6,
            Label = 7,
            HScrollBar = 8,
            VScrollBar = 9,
            TextBox = 10,
            NumericUpDown = 11,
            ProgressBar = 12,
            MenuStrip = 13,
            MenuStripButton = 14,
            MenuStripLabel = 15,
            MenuStripIcon = 16,
            MenuStripDropDownItem = 17,
            StatusStrip = 18,
            StatusStripProgressBar = 19,
            ToolStripMenu = 20,
            RichTextBox = 21, 
            ContextMenuStrip = 22,
            EventHandler = 1000
        }
    }
}