﻿using System;
using System.IO;
using System.Text;
using System.Reflection;
using System.Diagnostics;
using System.Threading;
using System.Windows.Forms;
using System.Collections;
using System.Runtime.InteropServices;
using Microsoft.Win32;

namespace Plywood {
    public class Interpreter : ICloneable {
        internal class definition : IDefinition {
            private object p_Value;
            private bool p_IsNumeric;
            private DefinitionType p_DefinitionType;
            private IList p_ArrayList;

            public definition() { 
                //Define default globals
                p_DefinitionType = DefinitionType.PlywodObject;
            }

            public string Name { get; set; }
            public object ValueInstance { get { return p_Value; } set {
                //Set the values
                p_Value = value;
            } }
            public string Value { get { 
                //Can we get the value
                if (p_DefinitionType == DefinitionType.CommonLanguageRuntimeObject) {
                    throw new Exception("Cannot get the string value from a CLR object");
                }
                return p_Value.ToString(); 
            } set { 
                //Can we set the value?
                if (p_DefinitionType == DefinitionType.CommonLanguageRuntimeObject) {
                    throw new Exception("Cannot set a string value to a CLR object");
                }
                
                //Set the type
                if (p_DefinitionType != Plywood.DefinitionType.ArrayObject) {
                    p_DefinitionType = Plywood.DefinitionType.PlywodObject;
                }
                p_Value = value;
            } }

            public bool IsNull { get { return Name == null; } }
            public bool IsNumeric { get {
                if (p_DefinitionType == Plywood.DefinitionType.CommonLanguageRuntimeObject) {
                    throw new Exception("Cannot get numerical flag from a CLR object");
                }
                return p_IsNumeric; 
            } set {
                //Can we change?
                if (p_DefinitionType == Plywood.DefinitionType.CommonLanguageRuntimeObject) {
                    throw new Exception("Cannot change the type of a CLR object");
                }
                
                p_IsNumeric = value;
                               
                //Change type
                p_DefinitionType = value ? DefinitionType.NumericalObject : DefinitionType.PlywodObject;
            } }
            public bool IsDefaultGenerated { get; set; }

            public DefinitionType DefinitionType { get { return p_DefinitionType; } }

            internal void setDef(DefinitionType type) { p_DefinitionType = type; }

            public IList ArrayList { get { return p_ArrayList; } set {
                //Can we set it
                if (p_DefinitionType == Plywood.DefinitionType.CommonLanguageRuntimeObject) {
                    throw new Exception("Cannot define an arry for a CLR object");
                }

                p_ArrayList = value;
                
                //Change the object type
                p_DefinitionType = value != null ? DefinitionType.ArrayObject : DefinitionType.PlywodObject;
            } }
            public bool Equals(IDefinition obj) { 
                //Can we compare?
                if (p_DefinitionType == Plywood.DefinitionType.CommonLanguageRuntimeObject) { 
                    throw new Exception("Cannot compare CLR objects");
                }

                //Same type?
                if (!(obj is IDefinition)) { return false; }

                //Get the object type
                IDefinition def = (IDefinition)obj;

                //Array?
                if ((ArrayList == null && def.ArrayList != null) ||
                    (ArrayList != null && def.ArrayList == null)) { return false; }

                //Check the array?
                if (ArrayList != null && def.ArrayList != null) {
                    //Same length?
                    if (ArrayList.Count != def.ArrayList.Count) { return false; }

                    //Compare the array elements
                    for (int c = 0; c < ArrayList.Count; c++) {
                        if (ArrayList[c] != def.ArrayList[c]) { return false; }
                    }

                    //There equal
                    return true;
                }

                //Just check it's value
                if (def.ValueInstance == null || ValueInstance == null) { return false; }
                return def.Value == Value;
            }
        }

        internal ITerminal[] p_Terminals = new ITerminal[0];
        internal IDefinition[] p_Defines = new IDefinition[0];
        internal IMethod[] p_Methods;
        internal string[] p_IgnoreLogicalPassList = new string[0];

        public IDefinition[] defines { get { return p_Defines; } }

        public Interpreter() { 
            //Let the local method handle this
            _constructor(true);
        }
        public Interpreter(bool initSubSystem) {
            //Let the local method handle this
            _constructor(initSubSystem);
        }

        private void _constructor(bool initSubSystem) {
            //Define the methods
            p_Methods = new GenericMethods(null).GetClrMethods();

            //Initialize defines
            Run(new MemoryStream(), true);

            //Define the default terminal
            AddTerminal(new WindowsTerminal());
            WindowsTerminal.Enabled = true;
        }

        public void ListMethodCalls(Stream stream, out IMethod[] methods, out string[] names) { 
            //Define the outputs temp defines to apply
            string[] namesTemp = new string[0];
            IMethod[] methodsTemp = new IMethod[0];

            //Buffer the stream position
            long strPos = stream.Position;

            //Define the state to pass the the execution engine
            RunState state = new RunState(this, stream);
            state.OnMethodInvoke += delegate(string fname, IMethod m, ref bool allowed) {
                //Block this function from being called
                allowed = false;

                //Does the method already exist in the array?
                foreach (string s in namesTemp) {
                    if (s == fname) { return; }
                }

                //Add the function to the arrays
                Array.Resize(ref namesTemp, namesTemp.Length + 1);
                Array.Resize(ref methodsTemp, methodsTemp.Length + 1);
                namesTemp[namesTemp.Length - 1] = fname;
                methodsTemp[namesTemp.Length - 1] = m;
            };

            //Run the script
            new Interpreter().Run(stream, state);

            //Pop the stream position back
            stream.Position = strPos;

            //Define the outputs
            names = namesTemp;
            methods = methodsTemp;
        }

        public object GetDefinition(string name) { 
            //Look for the value
            IDefinition def = new definition();
            bool found = false;
            foreach (IDefinition d in p_Defines) {
                if (d.Name == name) { def = d; found = true; break; }
            }

            //Found?
            if (!found) {
                throw new Exception("Definition was not found");
            }

            //Is it a numeric?
            if (def.IsNumeric) { 
                //Return double?
                if (def.Value.Contains(".")) { return Convert.ToDouble(def.Value); }
                //Just return a integer type
                return Convert.ToInt64(def.Value);
            }

            //Array?
            if (def.ArrayList != null) { return def.ArrayList; }

            //Return the string value
            return def.Value;
        }

        public string Run(Stream stream) { return Run(stream, false, new RunState(this, stream)); }
        public string Run(Stream stream, bool clearDefines) { return Run(stream, clearDefines, new RunState(this, stream)); }
        public string Run(Stream stream, RunState state) { return Run(stream, false, state); }
        public string Run(Stream stream, bool clearDefines, RunState state) {
            //Grab the methods
            charRange[] eRanges = null;
            string[] eLibNames, eRetTypes;
            string[] eMethods = getExternNames(stream, out eRanges, out eLibNames, out eRetTypes);
            charRange[] ranges = null;
            string[] methods = getMethodNames(stream, out ranges, eRanges);

            //Buffer the stream position
            long strPos = stream.Position;

            //Add the methods to the global array
            for (int c = 0; c < methods.Length; c++) {
                //Read the scope and begin code
                byte[] beginScript = new byte[ranges[c].Length];
                stream.Position = ranges[c].Start;
                stream.Read(beginScript, 0, beginScript.Length);
                string scope = "";
                while ((scope = skipScope(stream)) == null) { stream.Position++; }
                scope = remUnwantedCharacters(scope);
                                
                //Add the method
                AddMethod(new scriptMethod(scope, Encoding.ASCII.GetString(beginScript), methods[c], this));
            }
            
            //Add the external methods
            for (int c = 0; c < eMethods.Length; c++) { 
                //Read the line
                stream.Position = eRanges[c].Start;
                byte[] lRaw = new byte[eRanges[c].Length];
                stream.Read(lRaw, 0, lRaw.Length);

                //Add the method
                AddMethod(new externMethod(eMethods[c], eLibNames[c], eRetTypes[c], Encoding.ASCII.GetString(lRaw)));
            }

            //Pop the previous stream position back
            stream.Position = strPos;

            //Add the external method ranges to the method ranges
            foreach (charRange r in eRanges) {
                Array.Resize(ref ranges, ranges.Length + 1);
                ranges[ranges.Length - 1] = r;
            }

            //Shall we clear definitions?
            if (clearDefines) {
                p_Defines = new definition[0];
                //Add the default values
                foreach (PropertyInfo p in typeof(GenericMethods).GetProperties()) {
                    definition d = new definition();
                    try {
                        d.Name = p.Name;
                        d.Value = p.GetValue(null, null).ToString();
                        d.IsDefaultGenerated = true;
                        try {
                            Convert.ToInt64(d.Value);
                            d.IsNumeric = true;
                        }
                        catch { }                        
                    }
                    catch { continue; }
                    Array.Resize(ref p_Defines, p_Defines.Length + 1);
                    p_Defines[p_Defines.Length - 1] = d;
                }
            }

            //Loop through the stream
            while (stream.Position < stream.Length) { 
                //Dereference the current definition set to the state
                state.CurrentDefinition = new definition() { Name = null };
                state.ApplyDefinitionGlobal = false;

                //Skip over the comments and scopes
                while (!skipComment(stream)) ;
                skipScope(stream);
                if (stream.Position >= stream.Length) { break; }

                //Read the function name
                bool wasNewLine = false;
                string functionName = "";
                skipWhiteSpaces(stream);
                while (stream.Position < stream.Length) {
                    byte b = (byte)stream.ReadByte();
                    if (b == (byte)'\r') { continue; }
                    if (b == (byte)'=') { stream.Seek(-1, SeekOrigin.Current); break; }
                    if (b == (byte)'{') { wasNewLine = true; stream.Seek(-1, SeekOrigin.Current); break; }
                    if (b == (byte)' ') { break; }
                    if (b == (byte)'\n') { wasNewLine = true; break; }
                    functionName += (char)b;
                }

                //Is it a function?
                bool wasMethod = false;
                foreach (IMethod m in p_Methods) {
                    if (functionName.Split('(')[0] != m.Name) { continue; }
                    
                    //Get the rest of the line
                    if (!wasNewLine) { functionName += " "; }
                    while (stream.Position < stream.Length && !wasNewLine) {
                        byte b = (byte)stream.ReadByte();
                        if (b == (byte)'{') { stream.Seek(-1, SeekOrigin.Current); break; }
                        if (b == (byte)'\n') { break; }
                        functionName += (char)b;
                    }

                    //Get its value
                    getAssociatedValue(functionName, state);
                    wasMethod = true;
                    break;
                }
                if (wasMethod) { continue; }

                //Should it be a function?
                if (functionName.Contains("(")) {
                    throw new Exception("Invalid line");
                }

                //Define if it was a variable
                bool wasVar = false;

                //Is the function name a variable name?
                for(int c = 0; c < p_Defines.Length; c++) {
                    IDefinition d = p_Defines[c];
                    if (d.Name != functionName) { continue; }

                    //Set the current definition in the state
                    state.CurrentDefinition = d;

                    //Read the equal operator
                    while (stream.Position < stream.Length) {
                        byte b = (byte)stream.ReadByte();

                        //Equal sign?
                        if (b == (byte)'=') { 
                            //Read the value
                            bool num = false; 
                            string value = preProcessStringValue(readValue(stream, ' ', false, true, false), out num, state);
                                
                            if (!state.ApplyDefinitionGlobal) {
                                p_Defines[c].Value = value;
                                p_Defines[c].IsNumeric = num;
                            }
                            wasVar = true;

                            //Set the definition?
                            if (state.ApplyDefinitionGlobal) {
                                applyDef(state.CurrentDefinition);
                            }
                            break;
                        }

                        //No equal sign found?
                        if (b == (byte)'\n') {
                            throw new Exception("\"" + functionName + "\" is used incorrectly");
                        }
                    }
                    if (wasVar) { break; }
                }

                //Skip?
                if (wasVar) { continue; }

                //Return value?
                if (functionName == "ret") {
                    bool isNumeric = false;
                    return preProcessStringValue(readValue(stream, ' ', false, true, false), out isNumeric, state);
                }

                //Method?
                if (functionName == "method" || functionName == "extern") { 
                    //Valid?
                    bool valid = false;
                    charRange range = new charRange();
                    foreach (charRange r in ranges) {
                        if (stream.Position >= r.Start &&
                            stream.Position < r.Start + r.Length) {
                                valid = true;
                                range = r;
                                break; 
                        }
                    }
                    if (valid) {
                        //Skip over the scope
                        stream.Position = range.Start + range.Length;
                        skipScope(stream);
                        continue;
                    }
                }

                //Defining a value?
                if (functionName == "def") { 
                    //Read the name
                    string name = "";
                    bool newLineFound = false;
                    skipWhiteSpaces(stream);
                    while (stream.Position < stream.Length) {
                        byte b = (byte)stream.ReadByte();
                        if (b == (byte)' ') { break; }
                        if (b == (byte)'=') { stream.Seek(-1, SeekOrigin.Current); break; }
                        if (b == (byte)'\n') { newLineFound = true; break; }
                        name += (char)b;
                    }
                    name = remUnwantedCharacters(name);
                    
                    //Valid name?
                    if (!validName(name)) {
                        throw new Exception("Invalid variable \"" + name + "\" definition");
                    }


                    //Add the definition;
                    Array.Resize(ref p_Defines, p_Defines.Length + 1);
                    p_Defines[p_Defines.Length - 1] = new definition {
                        Name = name
                    };

                    //Set the state definition
                    state.CurrentDefinition = p_Defines[p_Defines.Length - 1];
                                       
                    //Read the value
                    while (stream.Position < stream.Length && !newLineFound) {
                        byte b = (byte)stream.ReadByte();
                        if (b == (byte)'=') {
                            bool num = false;

                            string str = readValue(stream, ' ', false, true, false);
                            p_Defines[p_Defines.Length - 1].Value = preProcessStringValue(str, out num, state);
                            p_Defines[p_Defines.Length - 1].IsNumeric = num;
                            break;
                        }                        
                        if (b == (byte)'\n') { break; }
                    }

                    //Set the definition value
                    if (state.ApplyDefinitionGlobal) { 
                        applyDef(state.CurrentDefinition); 
                    }
                    continue;
                }

                //Was it a blank line to be skipped over?
                if (functionName.Replace(" ", "").Length == 0) {
                    continue;
                }

                //This function was not processed
                throw new Exception("Invalid function \"" + functionName + "\"");
            }

            return "";
        }

        public void AddDefinition(string name, object value, DefinitionType type) { 
            //Define the definition to create
            definition d = new definition();

            //Is it a clr object?
            if (type == DefinitionType.CommonLanguageRuntimeObject) { 
                //Just add a CLR definition
                d.setDef(type);
                d.ValueInstance = value;
                AddDefinition(d); 
                return;
            }
                      
            //Array?
            if (value == typeof(IList)) {
                d.Value = "ARRAY_DEF";
                d.ArrayList = (IList)value;
                AddDefinition(d);
                return;
            }

            //Set the value
            d.Value = value.ToString();
            
            //Is it numerical?
            try {
                Convert.ToDouble(value);
                d.IsNumeric = true;
            }
            catch { }

            //Add the definition
            AddDefinition(d);
        }
        public void AddDefinition(IDefinition definition) {
            //Check the definition
            if (!validName(definition.Name)) {
                throw new Exception("Invalid name");
            }

            //Add the definition
            Array.Resize(ref p_Defines, p_Defines.Length + 1);
            p_Defines[p_Defines.Length - 1] = definition;
        }
        public void RemoveDefinition(string name) { 
            //Define the rebuild buffer
            IDefinition[] buffer = new IDefinition[p_Defines.Length - 1];
            int ai = 0;
            bool found = false;

            //Ignore the definition
            for (int c = 0; c < p_Defines.Length; c++) {
                if (p_Defines[c].Name == name) { found = true; continue; }
                buffer[ai] = p_Defines[c];
                ai++;
            }

            //Found?
            if (!found) {
                throw new Exception("Definition was not found");
            }

            //Set the global array to the rebuild buffer
            p_Defines = buffer;
        }

        public void AddTerminal(ITerminal terminal) { 
            //Already exist?
            foreach (ITerminal t in p_Terminals) {
                if (t == terminal) {
                    throw new Exception("Terminal instance already exists");
                }
            }

            //Add the terminal
            Array.Resize(ref p_Terminals, p_Terminals.Length + 1);
            p_Terminals[p_Terminals.Length - 1] = terminal;
        }
        public void RemoveTerminal(ITerminal terminal) { 
            //Remove the terminal
            ITerminal[] buffer = new ITerminal[p_Terminals.Length - 1];
            bool found = false;
            int ai = 0;
            for (int c = 0; c < p_Terminals.Length; c++) {
                if (p_Terminals[c] == terminal) { found = true; continue; }
                buffer[ai] = p_Terminals[c];
                ai++;
            }
            if (!found) {
                throw new Exception("Terminal does not exist");
            }
            p_Terminals = buffer;
        }

        public void AddMethod(IMethod method) { 
            //Does the method exist?
            foreach (IMethod m in p_Methods) {
                if (m.Name == method.Name) {
                    throw new Exception("Method already exists");
                }
            }

            //Add the method
            Array.Resize(ref p_Methods, p_Methods.Length + 1);
            p_Methods[p_Methods.Length - 1] = method;
        }
        public IMethod AddMethod(MethodInfo method) { 
            //Define the clr method
            IMethod m = new ClrMethod(method);

            //Add it and return it
            AddMethod(m);
            return m;
        }
        public void RemoveMethod(IMethod method) { 
            //Find the method
            bool found = false;
            foreach (IMethod m in p_Methods) {
                if (m == method) { found = true; break; }
            }
            if (!found) {
                throw new Exception("Method does not exist");
            }

            //Remove this object from the array
            IMethod[] buffer = new IMethod[p_Methods.Length - 1];
            int ai = 0;
            for (int c = 0; c < p_Methods.Length; c++) {
                if (p_Methods[c] == method) { continue; }
                buffer[ai] = p_Methods[c];
                ai++;
            }
            p_Methods = buffer;
        }
        public IMethod[] GetMethods(string name) { 
            //Define the return buffer 
            IMethod[] buffer = new IMethod[0];

            //Search for the names
            foreach (IMethod m in p_Methods) {
                if (m.Name == name) {
                    Array.Resize(ref buffer, buffer.Length + 1);
                    buffer[buffer.Length - 1] = m;
                }
            }

            //Return the buffer
            return buffer;
        }
        public IMethod GetMethod(string name) { 
            //Get the methods
            IMethod[] res = GetMethods(name);

            //Found?
            if (res.Length == 0) {
                throw new Exception("Method was not found");
            }

            //Return the firt entry
            return res[0];
        }
       
        public object Clone() { 
            //Define the return instance
            Interpreter buffer = new Interpreter();
            
            //Clone the definitions
            buffer.p_Defines = new IDefinition[p_Defines.Length];
            for (int c = 0; c < buffer.p_Defines.Length; c++) {
                //CLR object?
                if (p_Defines[c].DefinitionType == DefinitionType.CommonLanguageRuntimeObject) { 
                    //Just add the clr object
                    buffer.p_Defines[c] = new definition() {
                        ValueInstance = p_Defines[c].ValueInstance,
                        Name = p_Defines[c].Name
                    };
                    ((definition)buffer.p_Defines[c]).Name = p_Defines[c].Name;
                    continue;
                }

                buffer.p_Defines[c] = new definition { 
                    IsNumeric = p_Defines[c].IsNumeric,
                    IsDefaultGenerated = p_Defines[c].IsDefaultGenerated,
                    Name = p_Defines[c].Name,
                    ValueInstance = p_Defines[c].ValueInstance,
                    ArrayList = p_Defines[c].ArrayList != null ? p_Defines[c].ArrayList : null
                };
                ((definition)buffer.p_Defines[c]).setDef(p_Defines[c].DefinitionType);
            }

            //Define the methods and terminals
            buffer.p_Methods = p_Methods;
            buffer.p_Terminals = p_Terminals;


            //Return the buffer
            return buffer;
        }

        public ITerminal WindowsTerminal { get { 
            //Check
            if (p_Terminals.Length == 0 || p_Terminals[0].GetType() != typeof(WindowsTerminal)) {
                throw new Exception("Windows terminal not found");
            }

            //Return the terminal
            return p_Terminals[0];
        } }

        internal void addIgnoreLogicalPass(string methodName) { 
            //Add the method
            Array.Resize(ref p_IgnoreLogicalPassList, p_IgnoreLogicalPassList.Length + 1);
            p_IgnoreLogicalPassList[p_IgnoreLogicalPassList.Length - 1] = methodName;
        }
        internal object getClrType(string name) { 
            //Look up the name
            foreach (IDefinition d in p_Defines) {
                if (d.Name == name) { 
                    //Is it a CLR?
                    if (d.DefinitionType != DefinitionType.CommonLanguageRuntimeObject) {
                        throw new Exception("Definition is not a CLR object");
                    }

                    //Return the CLR object
                    return d.ValueInstance;
                }
            }

            //Not found
            throw new Exception("Definition was not found");
        }
        internal IDefinition getDefByName(string name) { 
            //Look for the definition
            name = name.Replace(" ", "");
            foreach (IDefinition d in p_Defines) {
                if (d.Name == name) { return d; }
            }

            //Not found 
            throw new Exception("Definition was not found");
        }
        internal void skipWhiteSpaces(Stream stream) { 
            //Ignore white spaces
            while (stream.ReadByte() == (byte)' ') ;

            //Go back 1 byte
            stream.Seek(-1, SeekOrigin.Current);
        }
        internal string readValue(string str, char endCharToken, bool allowFullString, bool readLine, bool stopReadOnString) { 
           //Let the local method handle this
            Stream st = createStrStream(str);
            string buffer = readValue(st, endCharToken, allowFullString, readLine, stopReadOnString);
            st.Close();
            return buffer;
        }
        internal string readValue(Stream stream, char endCharToken, bool allowFullString, bool readLine, bool stopReadOnString) { 
            //Let the local method handle this
            bool f = false;
            return readValue(stream, endCharToken, allowFullString, readLine, stopReadOnString, out f);
        }
        internal string readValue(Stream stream, char endCharToken, bool allowFullString, bool readLine, bool stopReadOnString, out bool wasNewLine) { 
            //Define the default outputs
            wasNewLine = false;
            
            //Define the return buffer
            string buffer = "";

            //Skip over white spaces
            skipWhiteSpaces(stream);
                                   
            //Add the characters to the buffer
            while (stream.Position < stream.Length) { 
                //Read the byte
                byte b = (byte)stream.ReadByte();

                //Ignore?
                if (b == (byte)'\r') {
                    continue;
                }

                //Scope break?
                if (b == (byte)'{') {
                    stream.Seek(-1, SeekOrigin.Current);
                    wasNewLine = true;
                    return buffer;
                }

                //Is this character a string char?
                if (b == (byte)'"' && allowFullString) { 
                    //Read the string
                    bool skipNextString = false;
                    string build = "";
                    while (stream.Position < stream.Length) {
                        b = (byte)stream.ReadByte();
                        if (b == (byte)'"' && !skipNextString) { break; }
                        if (b == (byte)'\\') { skipNextString = true; continue; }
                        skipNextString = false;
                        build += (char)b;
                    }
                    if (stopReadOnString) { return build; }
                    buffer += build;
                    continue;
                }

                //Read line?
                if (b == (byte)'\n' && readLine) {
                    wasNewLine = true;
                    return buffer;
                }

                //Found the end?
                if ((b == (byte)endCharToken || b == (byte)'\n') && !readLine) {
                    if (b != (byte)'\n') { stream.Seek(-1, SeekOrigin.Current); }
                    break; 
                }

                //Add the byte to the buffer
                buffer += (char)b;
            }

            //Return the buffer
            return buffer;
        }
        internal byte[] select(Stream stream, int length) { 
            //Define the return buffer
            byte[] buffer = new byte[length];

            //Read the bytes and push the stream position back
            stream.Seek(-stream.Read(buffer, 0, buffer.Length), SeekOrigin.Current);

            //Return the return buffer
            return buffer;
        }
        internal double getLongValue(string value) { 
            //Is it a declared value?
            foreach (IDefinition v in p_Defines) {
                if (v.Name == value && v.IsNumeric) {
                    return Convert.ToDouble(v.Value);
                }
            }

            //Return hex value?
            if (value[0] == '#') { 
                //Is this the only character?
                if (value.Length == 1) {
                    throw new Exception("Invalid hex value");
                }

                //Return the hex value
                return (double)NumericOperations.FromBinary(NumericOperations.FromHex(value.Substring(1, value.Length - 1)));
            }

            //Return binary?
            if (value[0] == 'b') { 
                //Is this the only character
                if (value.Length == 1) {
                    throw new Exception("Invalid binary value");
                }

                //Return the binary value
                return (double)NumericOperations.FromBinary(value.Substring(1, value.Length - 1));
            }

            //Just return the numerical value
            return Convert.ToDouble(value);
        }
        internal string getAssociatedValue(string value, RunState state) {
            //Is it a definition?
            foreach (IDefinition d in p_Defines) {
                if (d.Name == value.Replace(" ", "")) { return d.Value; }
            }
            
            //Is it a function?
            if (value.Contains("(") && value.Contains(")") && value.Split('(')[0].Length != 0) {
                //Get the function name
                string functionName = value.Split('(')[0].Replace(" ", "");
                               
                //Get the string which contains the arguments
                string argStringRAW = "";
                for (int c = 0; c < value.Length; c++) {
                    if (value[c] == '(') {
                        c++;
                        int length = value.Length - c;
                        int s = c;
                        while (value[c + length - 1] != ')') { length--; } length--;
                        argStringRAW = value.Substring(s, length);
                        break;
                    }
                }

                //Get the argument data
                string[] argData = new string[0];
                string build = "";
                for (int c = 0; c < argStringRAW.Length; c++) { 
                    //Is it a string?
                    if (argStringRAW[c] == '"') { 
                        //Skip over this character
                        build += "\"";
                        c++;
                        bool wasLastCharSkip = false;
                        while (c < argStringRAW.Length) { 
                            //Ignore next string character?
                            if (argStringRAW[c] == '\\') { wasLastCharSkip = true; build += "\\"; c++; continue; }
                            if (argStringRAW[c] == '"' && !wasLastCharSkip) { build += "\""; break; }
                            wasLastCharSkip = false;
                            build += argStringRAW[c];
                            c++;
                        }
                        continue;
                    }

                    //Is it a open brace?
                    if (argStringRAW[c] == '(') {
                        //Skip over all the brackets
                        int bOpenCount = 0;
                        int bCloseCount = 0;
                        for (int x = c; x < argStringRAW.Length; x++, c++) {
                            if (argStringRAW[c] == '(') { bOpenCount++; }
                            if (argStringRAW[x] == ')') { 
                                bCloseCount++;
                                if (bCloseCount == bOpenCount) {
                                    build += ')';
                                    break;
                                }
                            }
                            build += argStringRAW[c];
                        }
                        continue;
                    }

                    //Is it a argument split character
                    if (argStringRAW[c] == ',') { 
                        //Is it valid?
                        if (build.Replace(" ", "").Length == 0) {
                            throw new Exception("No value was defined");
                        }

                        //Add the build data to the argument list
                        Array.Resize(ref argData, argData.Length + 1);
                        argData[argData.Length - 1] = build;
                        build = "";
                        continue;
                    }

                    //Just add this character to the build
                    build += argStringRAW[c];
                }

                //Clean up
                if (build.Replace(" ", "").Length != 0) {
                    Array.Resize(ref argData, argData.Length + 1);
                    argData[argData.Length - 1] = build;
                }

                //Grab all the values to pass to the function
                object[] values = new object[argData.Length];
                for (int c = 0; c < argData.Length; c++) { 
                    //Get the value                    
                    values[c] = argData[c];
                }

                //Define if the function name was found and threw
                bool functionFound = false;
                bool functionThrew = false;

                //Find the function
                foreach (IMethod m in p_Methods) {
                    if (m.Name == functionName) {
                        //Grab the method arguments
                        Type[] argTypes = m.ArgumentTypes;
                        
                        //Define the values again for selection
                        object[] selectArg = (object[])values.Clone();
                                                
                        //Is the last type an array? (for param)
                        bool wasParam = false;
                        if (argTypes.Length != 0 &&
                            argTypes[argTypes.Length - 1] == typeof(object[]) &&
                            values.Length >= argTypes.Length - 1) {
                                wasParam = true;
                                //Trim the rest of the arguments into an object array
                                object[] o = new object[values.Length - argTypes.Length + 1];
                                
                                //Push the rest of the selected arguments into the array
                                for (int c = 0; c < o.Length; c++) {
                                    o[c] = getValueFromRaw(values[values.Length - c - 1].ToString(), state);
                                }

                                //Reverse the array 
                                object[] r = new object[o.Length];
                                for (int c = 0; c < r.Length; c++) {
                                    r[c] = o[o.Length - c - 1];
                                }

                                //Set the entry
                                Array.Resize(ref selectArg, selectArg.Length - o.Length + 1);
                                selectArg[selectArg.Length - 1] = r;
                        }

                        //Get the logical values
                        if (!wasParam) {
                            for (int c = 0; c < selectArg.Length; c++) {
                                if (!m.IgnoreLogicalArgument(c)) {
                                    selectArg[c] = getValueFromRaw(selectArg[c].ToString(), state); 
                                }
                            }
                        }

                        //Check them
                        if (argTypes.Length != selectArg.Length) { continue; }                                              
                        bool valid = true;
                        for (int c = 0; c < argTypes.Length; c++) {
                            if (argTypes[c] == typeof(object[]) && c == argTypes.Length - 1) { continue; }
                            Type a = selectArg[c].GetType();
                            Type b = argTypes[c];
                            bool na = a is long || a is double;
                            bool nb = b is long || b is double;
                            if (b == typeof(object[])) { continue; }
                            if (b == typeof(string)) {
                                selectArg[c] = selectArg[c].ToString();
                                continue; 
                            }                            
                            if (nb && !na) {
                                valid = false;
                                break;
                            }
                            if (b == typeof(long)) { selectArg[c] = Convert.ToInt64(selectArg[c]); }
                            if (b == typeof(double)) { selectArg[c] = Convert.ToDouble(selectArg[c]); }
                        }                        
                        if (!valid) { continue; }

                        //We have found the function
                        functionFound = true;
                        
                        try { 
                            //Are we allowed to invoke this function
                            bool allowed = false;
                            state.invokeMethod(m.Name, m, ref allowed);

                            //Not allowed? just return the default value
                            if (!allowed) { return "0"; }

                            //Invoke
                            object ret =
                                    (m.GetType() == typeof(GenericMethod) ?
                                        m.Invoke(state.Functions, selectArg) :
                                        m.Invoke(selectArg));

                            //Return bool value?
                            if (ret != null && ret is bool) {
                                return ret.ToString().ToLower();
                            }

                            //Return it's return value?
                            return ret == null ? "" : ret.ToString();
                        }
                        catch {
                            //Function threw
                            functionThrew = true;
                        }
                    }
                }

                //Was the function found?
                if (functionFound) {
                    //Was there an error in the function?
                    if (functionThrew) {
                        throw new Exception("An error occured internally for function \"" + functionName + "\"");
                    }
                    throw new Exception("Invalid arguments where passed to function \"" + functionName + "\"");
                }

                //Method not found
                throw new Exception("Function \"" + functionName + "\" not found");
            }

            //Throw
            throw new Exception("Invalid");
        }
        internal Stream createStrStream(string str) { 
            //Define the stream
            Stream buffer = new MemoryStream();

            //Write the string
            buffer.Write(Encoding.ASCII.GetBytes(str), 0, str.Length);
            buffer.Position = 0;

            //Return the stream 
            return buffer;
        }
        internal void applyDef(IDefinition def) { 
            //Look for the definition
            for (int c = 0; c < p_Defines.Length; c++) {
                if (p_Defines[c].Name == def.Name) { 
                    //Set all the values from the arguments to the found definition
                    p_Defines[c] = def;
                    return;
                }
            }

            //Not found
            throw new Exception("Definition was not found");
        }
        internal void applyDefinitionChanges(Interpreter int1, Interpreter int2) { 
            //Loop through every definition in the first interpreter
            for (int c = 0; c < int1.p_Defines.Length; c++) { 
                //Has the value changed?
                foreach (IDefinition d in int2.p_Defines) {
                    //Ignore?
                    if (d.DefinitionType == DefinitionType.CommonLanguageRuntimeObject ||
                        int1.p_Defines[c].DefinitionType == DefinitionType.CommonLanguageRuntimeObject) { 
                        continue; 
                    }
                    if (d.Name == int1.p_Defines[c].Name) {
                        if (!d.Equals(int1.p_Defines[c]) && d.ValueInstance != null) {
                            int1.p_Defines[c] = d;
                        }
                        break;
                    }
                }
            }
        }

        internal bool skipComment(Stream stream) {
            //Skip over white spaces
            while (select(stream, 1)[0] == (byte)' ') { stream.Seek(1, SeekOrigin.Current); }
            
            //Select the first 2 bytes for processing the comment
            byte[] commentSelect = select(stream, 2);

            //Line comment?
            if (commentSelect[0] == (byte)'/' && commentSelect[1] == (byte)'/') {
                //Skip over the whole line
                while (stream.Position < stream.Length && stream.ReadByte() != '\n') ;
                return false;
            }

            //Block comment?
            if (commentSelect[0] == (byte)'/' && commentSelect[1] == (byte)'*') {
                while (stream.Position < stream.Length) {
                    //End block quote?
                    commentSelect = select(stream, 2);
                    if (commentSelect[0] == '*' && commentSelect[1] == '/') {
                        stream.Seek(2, SeekOrigin.Current);
                        break;
                    }

                    //Skip over this byte
                    stream.Seek(1, SeekOrigin.Current);
                }
                return false;
            }
            return true;
        }
        internal string skipScope(Stream stream) {
            //Is the first byte a scope?
            if (select(stream, 1)[0] != (byte)'{') { return null; }

            //Find the last scope character
            int sCount = 0;
            int cCount = 0;
            string buffer = "";
            while (stream.Position < stream.Length) {
                byte r = (byte)stream.ReadByte();
                if (r == (byte)'{') {
                    if (sCount != 0) { buffer += (char)r; }
                    sCount++; continue; }
                if (r == (byte)'}') {
                    cCount++;
                    if (cCount == sCount) { break; }
                }
                buffer += (char)r;
            }

            //Return the buffer
            return buffer;
        }
        internal void skipString(Stream stream) { 
            //Loop through the stream
            bool skipNextStr = false;
            while (stream.Position < stream.Length) { 
                //Read the byte
                byte b = (byte)stream.ReadByte();

                if (b == (char)'"' && !skipNextStr) { break; }
                if (b == (char)'\\') { skipNextStr = true; continue; }
                skipNextStr = false;
            }
        }
        internal string skipString(string value, ref int position) { 
            //Skip
            string buffer = "";
            bool skipStr = false;
            while (position < value.Length) {
                if (value[position] == '\\') { skipStr = true; position++; continue; }
                if (value[position] == '"' && !skipStr) { return buffer; }
                skipStr = false;
                buffer += value[position];
                position++;
            }
            return null;
        }

        internal bool validName(string name) { 
            //Name already exists?
            foreach (IDefinition d in p_Defines) {
                if (d.Name == name) { return false; }
            }

            //Is it a syntax token?
            if (name == "def" || name == "ret") { return false; }

            //Is there a function with the same name?
            foreach (IMethod m in p_Methods) {
                if (m.Name == name) { return false; }
            }

            //Does the name begin with a number?
            try {
                Convert.ToSByte(name[0].ToString());
                return false;
            }
            catch { }

            //Contains invalid characters?
            string invalidCharacters = "!\"£$%^&*()_+-={}[]:;@'~#<,>.?/|\\";
            foreach (char c in name) {
                foreach (char x in invalidCharacters) {
                    if (x == c) { return false; }
                }
            }

            //It's valid
            return true;
        }
        internal string remUnwantedCharacters(string value) { 
            //Replace the unwanted characters
            string unwant = "\r\t";
            foreach (char c in unwant) {
                value = value.Replace(c.ToString(), "");
            }

            //Just return the value
            return value;
        }

        internal object getValueFromRaw(string rawValue, RunState state) { 
            //Is the value a definition?
            foreach (IDefinition d in p_Defines) {
                if (rawValue.Replace(" ", "") == d.Name) {
                    return d.ValueInstance;
                }
            }

            //Process the value
            bool numeric = false;
            rawValue = preProcessStringValue(rawValue, out numeric, state);

            //Return a numeric type?
            if (numeric) { 
                //Double?
                if (rawValue.Contains(".")) { return Convert.ToDouble(rawValue); }
                
                //Just return an integer
                try { return Convert.ToInt64(rawValue); }
                catch { numeric = false; }
            }

            //Is it boolean
            if (rawValue == "false" || rawValue == "true") {
                return rawValue == "true";
            }

            //Just return the value (as a string)
            return rawValue;
        }
        internal bool getBoolMultiLogic(string value, RunState state) {
            //Is it a multiple logic expression?
            bool isOr = false;
            if (!isMultiLogicalExpression(value, out isOr)) { 
                //Just perform the single logic expression
                return getBool(value, state);
            }

            //Define the sub-logic operations to perform
            string[] subExpressions = new string[0];
            string build = "";

            //Loop through the value
            for (int c = 0; c < value.Length; c++) { 
                //String?
                if (value[c] == '"') {
                    build += "\"";
                    build += skipString(value, ref c) + "\"";
                    continue;
                }
                //Brace?
                if (value[c] == '(') {
                    string b = getScopeEndIndexWithStr(value, c, '(', ')');
                    c += b.Length + 1;
                    build += "(" + b + ")";
                    continue;
                }
                //Is it a logical operator?
                string sub = "";
                try { sub = value.Substring(c, 2); } catch { }
                bool isOp = sub == "&&" || sub == "||";
                bool or = sub == "||";
                if (!isOp) {
                    build += value[c];
                    continue; 
                } 

                //Check the operator
                if (or && !isOr || isOr && !or) {
                    throw new Exception("Invalid logical operation");
                }

                //Skip over this operator
                c++;

                //Add the build to the logical cache
                Array.Resize(ref subExpressions, subExpressions.Length + 1);
                subExpressions[subExpressions.Length - 1] = build;
                build = "";
            }

            //Clean up
            if (build.Length != 0) {
                Array.Resize(ref subExpressions, subExpressions.Length + 1);
                subExpressions[subExpressions.Length - 1] = build;
                build = "";
            }

            //Perform all the logic 
            bool[] outputs = new bool[subExpressions.Length];
            int trueCount = 0;
            for (int c = 0; c < outputs.Length; c++) { 
                outputs[c] = getBool(subExpressions[c], state); 
                //True?
                if (outputs[c]) { trueCount++; }
            }

            //Is it or and at least 1 expression true?
            if (isOr) { return trueCount != 0; }
            
            //Are all outputs true?
            return (trueCount == outputs.Length);
        }
        internal bool getBool(string value, RunState state) {
            //Attempt to convert the value to a boolean?
            if (!isLogicalExpression(value)) {
                //Preprocess the value
                bool numeric = false;
                bool invertBool = false;
                if (value.StartsWith("!")) { invertBool = true; value = value.Substring(1); }
                try { value = preProcessStringValue(value, out numeric, state); } catch { }
                if (numeric) { return false; }

                //Remove braces?
                if (value.StartsWith("(") && value.EndsWith(")")) {
                    value = value.Substring(1);
                    value = value.Substring(0, value.Length - 1);
                                       
                    //Reprocess
                    value = preProcessStringValue(value, out numeric, state);
                    if (numeric) { return false; }
                }

                //Just convert the value to a boolean                
                return invertBool ? !Convert.ToBoolean(value) :
                                     Convert.ToBoolean(value);
            }

            //Remove the whitespaces at the beginning and end
            int offset = 0;
            while (offset < value.Length) { if (value[offset] != ' ') { break; } offset++; }
            value = value.Substring(offset);
            offset = 0;
            while (offset < value.Length) { if (value[value.Length - offset - 1] != ' ') { break; } offset++; }
            value = value.Substring(0, value.Length - offset);

            //Define the logical operation
            logicOperator logicOperator = Interpreter.logicOperator.UNKNOWN;
            string logicOperatorStr = "";
            int logicOpPosition = 0;

            //Read the first side of the logic
            string leftValue = "";
            bool leftValueIsInt = false;
            for (int c = 0; c < value.Length; c++) {
                if (value[c] == '(') {
                    string b = getScopeEndIndexWithStr(value, c, '(', ')');
                    leftValue += "(" + b + ")";
                    c += b.Length + 1;
                    continue;
                }
                logicOperator = isLogicOp(value, c, out logicOperatorStr);
                if (logicOperator != logicOperator.UNKNOWN) {
                    logicOpPosition = c;
                    break; 
                }
                leftValue += value[c];
            }
            if (logicOperator == Interpreter.logicOperator.UNKNOWN) {
                throw new Exception("Invalid logical operator");
            }
            leftValue = preProcessStringValue(leftValue, out leftValueIsInt, state);

            //Read the right side of the logic
            string rightValue = "";
            bool rightValueIsInt = false;
            for (int c = logicOpPosition + logicOperatorStr.Length; c < value.Length; c++) {
                if (value[c] == '(') {
                    string b = getScopeEndIndexWithStr(value, c, '(', ')');
                    c += b.Length + 1;
                    rightValue += "(" + b + ")";
                    continue;
                }               
                rightValue += value[c];
            }
            rightValue = preProcessStringValue(rightValue, out rightValueIsInt, state);

            //Convert either to a string?
            if (!leftValueIsInt || !rightValueIsInt) {
                leftValueIsInt = false;
                rightValueIsInt = false;
            }

            //Check the operator
            if (logicOperator != Interpreter.logicOperator.EQUAL && logicOperator != Interpreter.logicOperator.NOT_EQUAL &&
                !leftValueIsInt && !rightValueIsInt) {
                    throw new Exception("Cannot perform \"" + logicOperatorStr + "\" on strings");
            }

            //Do the logic
            switch (logicOperator) { 
                case Interpreter.logicOperator.EQUAL:
                    return leftValue == rightValue;

                case Interpreter.logicOperator.NOT_EQUAL:
                    return rightValue != leftValue;

                case Interpreter.logicOperator.MORE:
                    return Convert.ToDecimal(leftValue) > Convert.ToDecimal(rightValue);

                case Interpreter.logicOperator.LESS:
                    return Convert.ToDecimal(leftValue) < Convert.ToDecimal(rightValue);
                
                case Interpreter.logicOperator.MORE_EQUAL:
                    return Convert.ToDecimal(leftValue) >= Convert.ToDecimal(rightValue);
                
                case Interpreter.logicOperator.LESS_EQUAL:
                    return Convert.ToDecimal(leftValue) <= Convert.ToDecimal(rightValue);
            }

            //Throw
            throw new Exception("Invalid statement");
        }
        internal bool isMultiLogicalExpression(string value) { 
            //Let the local method handle this
            bool o = false;
            return isMultiLogicalExpression(value, out o);
        }
        internal bool isMultiLogicalExpression(string value, out bool isOr) { 
            //Define default outputs
            isOr = false;

            //Look for && and || operators
            for (int c = 0; c < value.Length - 1; c++) { 
                //String?
                if (value[c] == '"') {
                    c++;
                    skipString(value, ref c);
                    continue;
                }
                //Brace?
                if (value[c] == '(') {
                    c = getScopeEndIndex(value, c, '(', ')');
                    continue;
                }

                //Get the 2 character substring from this position
                string sub = value.Substring(c, 2);

                //Or operator?
                if (sub == "||") { isOr = true; return true; }
                //And operator?
                if (sub == "&&") { return true; }
            }

            //Not found 
            return false;
        }
        internal bool isLogicalExpression(string value) { 
            //Look for a logical operator
            for (int c = 0; c < value.Length; c++) {
                if (value[c] == '"') {
                    c++;
                    skipString(value, ref c);
                    continue;
                }
                if (value[c] == '(') {
                    c = getScopeEndIndex(value, c, '(', ')');
                    continue;
                }

                if (isLogicOp(value, c) != logicOperator.UNKNOWN) {
                    return true;
                }
            }

            //No occurances of logical operators where found
            return false;
        }
        internal logicOperator isLogicOp(string str, int position) { 
            //Let the local method handle this
            string t;
            return isLogicOp(str, position, out t);
        }
        internal logicOperator isLogicOp(string str, int position, out string operationStr) { 
            //Define the logic operator map
            string[] logicMap = { 
                                    "==", "!=", 
                                    ">=", "<=", ">", "<"
                                };

            //Loop through the logic map
            for (int c = 0; c < logicMap.Length; c++) { 
                //Ignore?
                if (position + logicMap[c].Length > str.Length) { continue; }

                //Read the appropriate string to compare
                string cmp = str.Substring(position, logicMap[c].Length);

                //Are they equal?
                if (cmp == logicMap[c]) { 
                    //Return the enum flag for this type
                    operationStr = logicMap[c];
                    return (logicOperator)c;
                }
            }

            //Not found
            operationStr = null;
            return logicOperator.UNKNOWN;
        }
        internal string preProcessStringValue(string value, out bool isNumerical, RunState state) {
            //Define defaults
            isNumerical = false;

            //Remove unwanted characters
            value = remUnwantedCharacters(value);

            //Trim the value
            int trimOffset = 0;
            while (value[trimOffset] == ' ') { trimOffset++; }
            value = value.Substring(trimOffset);
            trimOffset = 0;
            while (value[value.Length - trimOffset - 1] == ' ') { trimOffset++; }
            value = value.Substring(0, value.Length - trimOffset);

            //Boolean value?
            if (value == "true" || value == "false") {
                return value;
            }

            //Is it a logical expression?
            if (isLogicalExpression(value) || isMultiLogicalExpression(value)) {
                return getBoolMultiLogic(value, state).ToString().ToLower();
            }

                        
            //Define the values that have been passed
            string[] varPasses = new string[0];
            string build = "";
            int opCount = 0;

            //Loop through every character
            for (int c = 0; c < value.Length; c++) {
                //String?
                if (value[c] == '"') {
                    c++;
                    build += "\"" + skipString(value, ref c) + "\"";
                    continue;
                }

                //& operator?
                if (value[c] == '&') { 
                    //Add the build to the array?
                    if (build.Length != 0) {
                        Array.Resize(ref varPasses, varPasses.Length + 1);
                        varPasses[varPasses.Length - 1] = build;
                        build = "";
                    }

                    //Skip over blank characters
                    while (c < value.Length && value[c] == ' ') { c++; }

                    //Incriment the operator count
                    opCount++;
                    continue;
                }

                //Just add the character to the buffer
                if (value[c] == ' ') { continue; }
                build += value[c];
            }

            //Clean up
            if (build.Length != 0) {
                Array.Resize(ref varPasses, varPasses.Length + 1);
                varPasses[varPasses.Length - 1] = build;
                build = "";
            }

            //Check
            if (opCount + 1 != varPasses.Length) {
                throw new Exception("Invalid value");
            }

            //Define the return string
            int numericCount = 0;
            string buffer = "";
            addAppropriateValue(varPasses[0], ref numericCount, ref buffer, state);

            //Fill in the values
            for (int c = 0; c < opCount; c++) {
                addAppropriateValue(varPasses[c + 1], ref numericCount, ref buffer, state);
            }

            //Numeric?
            if (numericCount == varPasses.Length) {
                isNumerical = true;
            }

            //Return the buffer
            return buffer;
        }
        internal double calculate(string calc) { 
            //Replace unwanted characters
            calc = calc.Replace(" ", "");

            //Just return 0?
            if (calc.Length == 0) {
                return 0;
            }

            //Calculate the values in the brackets
            string calcRebuild = "";
            for (int c = 0; c < calc.Length; c++) {
                if (calc[c] != '(') { calcRebuild += calc[c]; continue; }
                int eIndex = getScopeEndIndex(calc, c, '(', ')');
                calcRebuild += calculate(calc.Substring(c + 1, eIndex - c - 1));
                c = eIndex;
            }
            calc = calcRebuild;

            //Define the operation characters and numeric values
            char[] opChars = new char[0];
            double[] numeric = new double[0];

            //Define the buffer for building values
            string valBuffer = "";
            
            //Loop through the calculation characters
            for (int c = 0; c < calc.Length; c++) { 
                //Is this character a math operation?
                if (isMathOp(calc[c])) { 
                    //To early to be a math operator?
                    if (valBuffer.Length == 0 && calc[c] != '-') { throw new Exception("Invalid calculation"); }
                    if (valBuffer.Length == 0) {
                        valBuffer += calc[c];
                        continue;
                    }

                    //Push this character into the operation array
                    Array.Resize(ref opChars, opChars.Length + 1);
                    opChars[opChars.Length - 1] = calc[c];

                    //Add the numeric value from the buffer
                    Array.Resize(ref numeric, numeric.Length + 1);
                    numeric[numeric.Length - 1] = getLongValue(valBuffer);
                    valBuffer = "";
                    continue;
                }

                //Add the character to the buffer
                valBuffer += calc[c];
            }

            //Clean up the calculation parse
            if (valBuffer.Length != 0) {
                Array.Resize(ref numeric, numeric.Length + 1);
                numeric[numeric.Length - 1] = getLongValue(valBuffer);
                valBuffer = "";
            }

            //Check the arrays
            if (opChars.Length + 1 != numeric.Length) {
                throw new Exception("Invalid calculation");
            }

            //Peform the calculation
            double buffer = numeric[0];
            for (int c = 0; c < opChars.Length; c++) {
                double val = numeric[c + 1];
                switch (opChars[c]) {
                    case '+': buffer += val; break;
                    case '-': buffer -= val; break;
                    case '*': buffer *= val; break;
                    case '/': buffer /= val; break;
                    case '^': buffer = (double)((long)buffer ^ (long)val); break;
                    case '%': buffer %= val; break;
                }
            }

            //Return the calculation result
            return buffer;
        }
        internal void addAppropriateValue(string val, ref int numericCount, ref string buffer, RunState state) {            
            //Remove whitespaces at the beginning
            int wOffset = 0;
            while (val[wOffset] == ' ') { wOffset++; }
            val = val.Substring(wOffset);

            //Define value to work with
            string wVal = readValue(val, ' ', true, true, false);
            
            //Attempt to calculate outright
            try {
                buffer += calculate(val.Replace(" ", ""));
                numericCount++;
                return;
            }
            catch { }

            //Do we know this is a function? (to also stop calculation processing)
            bool ignoreCalcProcess = false;
            if (wVal.Contains("(")) {
                try {
                    if (getScopeEndIndex(wVal, 0, '(', ')') == wVal.Length) {
                        foreach (IMethod m in p_Methods) {
                            if (m.Name == wVal.Split('(')[0]) {
                                ignoreCalcProcess = true;
                                break;
                            }
                        }
                    }
                }
                catch {
                }               
            }

            //Create a string stream
            Stream valueStream = createStrStream(val);

            //Define if it is a computation
            bool isComp = false;
            
            //Define the calculation stack and function name buffer
            string calcStack = "";
            string calcFBuffer = "";

            //Loop through the string stream for any calculation operations
            while (valueStream.Position < valueStream.Length && !ignoreCalcProcess) { 
                //Read the byte
                byte b = (byte)valueStream.ReadByte();

                //Is it a string?
                if (b == (byte)'"') { break; }

                //Brace?
                if (b == (byte)'(') { 
                    //Skip over it?
                    if (!isComp) {
                        valueStream.Position = getScopeEndIndex(val, (int)valueStream.Position - 1, '(', ')');
                        continue;
                    }

                    //Get the string that is in between the braces
                    int braceStart = (int)valueStream.Position;
                    int braceLength = getScopeEndIndex(val, (int)valueStream.Position - 1, '(', ')');
                    valueStream.Position = braceLength + 1;
                    braceLength = braceLength - braceStart;
                    string bRaw = val.Substring(braceStart, braceLength);
                    
                    //Is it function braces?
                    bool func = false;
                    foreach (IMethod m in p_Methods) {
                        if (m.Name == calcFBuffer) { func = true; break; }
                    }
                    if (func) { 
                        //Remove the function name from the calculation buffer
                        calcStack = calcStack.Substring(0, calcStack.Length - calcFBuffer.Length);

                        //Invoke the function and add the result to the calculation stack
                        bool numeric = false;
                        string v = preProcessStringValue(
                                        calcFBuffer + "(" + bRaw + ")", out numeric, state);
                        if (numeric) {
                            calcStack += v;
                        }
                        calcFBuffer = "";
                        continue;
                    }
                    
                    //Process what is inside the braces and add the result to the calculation buffer
                    if (braceLength != 0) {
                        bool isNumeric = false;
                        string v = preProcessStringValue(bRaw, out isNumeric, state);

                        //Add to the calculation stack?
                        if (isNumeric) { 
                            calcStack += v;
                            continue;
                        }
                    }
                    calcFBuffer = "";
                    continue;
                }
                
                //Calculation operator?
                if (isMathOp((char)b)) {
                    if (!isComp) {
                        valueStream.Position = 0;
                        isComp = true;
                    }
                    else {
                        calcStack += (char)b;
                    }
                    calcFBuffer = "";
                    continue;
                }

                //Add to the calculation stack?
                if (isComp) { calcStack += (char)b; }
                if (b != (byte)' ') { calcFBuffer += (char)b; }
            }

            //Calculate?
            if (isComp) {
                buffer += calculate(calcStack);
                numericCount++;
                return;
            }

            //Attempt to get the value associated with this
            try {
                string v = getAssociatedValue(val, state);
                try { getLongValue(v); numericCount++; }
                catch { }
                buffer += v;
                return;
            }
            catch { }
            
            //Was it a string char?
            if (val.Contains("\"")) {
                //Is there any characters behind this?
                val = val.Replace(" ", "");
                if (val.IndexOf('"') != 0) {
                    throw new Exception("Invalid value");
                }
            }

            //Just add the string to the buffer
            buffer += wVal;
        }
        internal int getScopeEndIndex(string value, int position, char open, char close) {
            //Define the open and close count
            int oCount = 0;
            int cCount = 0;
            
            //Loop through the string
            for (int c = position; c < value.Length; c++) { 
                //Is it a string char?
                if (value[c] == '"') {
                    c++;
                    skipString(value, ref c);
                    continue;
                }

                //Open character?
                if (value[c] == open) { oCount++; }
                
                //Close count?
                if (value[c] == close) {
                    cCount++;
                    //Return this index?
                    if (oCount == cCount) {
                        return c;
                    }
                }
            }

            //Throw
            throw new Exception("End of scope cannot be found");
        }
        internal string getScopeEndIndexWithStr(string value, int position, char open, char close) { 
            //Find the initial location of the character
            int start = position;
            while (start < value.Length && value[start] != open) { start++; }
            
            //Get the end index
            int eIndex = getScopeEndIndex(value, start, open, close);
            eIndex++;

            //Return the string in the middle
            return value.Substring(start + 1, eIndex - position - 2);
        }
        internal bool isMathOp(char character) { 
            //Check for math operation
            return character == '+' || character == '-' || character == '*' || character == '/' ||
                   character == '^' || character == '%';
        }
        internal bool hasScope(Stream stream) { 
            //Buffer the stream position
            long position = stream.Position;

            //Loop through the stream
            while (stream.Position < stream.Length) { 
                //Read the next byte 
                byte b = (byte)stream.ReadByte();

                //Ignore?
                if (b == (byte)' ' || b == (byte)'\n') { continue; }
                
                //Found a scope?
                if (b == (byte)'{') {
                    stream.Position = position;
                    return true; }

                //Not found
                stream.Position = position;
                return false;
            }

            //Not found
            stream.Position = position;
            return false;
        }
        internal string[] getExternNames(Stream str, out charRange[] ranges, out string[] dllName, out string[] retType) { 
            //Define the return buffer
            string[] buffer = new string[0];
            
            //Define the default outputs
            ranges = new charRange[0];
            dllName = new string[0];
            retType = new string[0];            

            //Buffer the stream position and seek to the beginning of the stream
            long position = str.Position;
            str.Seek(0, SeekOrigin.Begin);
            
            //Loop through the stream 
            while (str.Position < str.Length) { 
                //Read the entire line
                long pos = str.Position;
                string line = "";
                while (str.Position < str.Length) {
                    byte b = (byte)str.ReadByte();
                    if (b == (byte)'\n') { break; }
                    line += (char)b;
                }

                //Define the length of the line
                long lineLength = line.Length;
                
                //Replace unwanted characters
                line = remUnwantedCharacters(line);
                if (line.Replace(" ", "").Length == 0) { continue; }
                int offset = 0;
                while (line[offset] == ' ') { offset++; }
                line = line.Substring(offset);
                offset = 0;
                while (line[line.Length - offset - 1] == ' ') { offset++; }
                line = line.Substring(0, line.Length - offset);
                
                //Process this line?
                if (!line.StartsWith("extern ") || !line.EndsWith(")")) { continue; }

                //Rebuild the line
                string[] split = line.Split(' ');
                line = "";
                foreach (string s in split) {
                    if (s.Replace(" ", "").Length == 0) { continue; }
                    line += s + " ";
                }
                if (line.Length != 0) { line = line.Substring(0, line.Length - 1); }

                //Read the dll name
                int lPos = 0;
                while (line[lPos] != '"') { lPos++; } lPos++;
                string libName = skipString(line, ref lPos);
                lPos++;

                //Read the mehod return type
                string rType = "";
                while (line[lPos] == ' ') { lPos++; }
                while (line[lPos] != ' ') {
                    rType += line[lPos];
                    lPos++;
                }

                //Read the name
                string methodName = "";
                while (line[lPos] == ' ') { lPos++; }
                while (line[lPos] != ' ' && line[lPos] != '(') {
                    methodName += line[lPos];
                    lPos++;
                }

                //Add the entry
                Array.Resize(ref dllName, dllName.Length + 1);
                Array.Resize(ref retType, retType.Length + 1);
                Array.Resize(ref buffer, buffer.Length + 1);
                dllName[dllName.Length - 1] = libName;
                retType[retType.Length - 1] = rType;
                buffer[buffer.Length - 1] = methodName;

                //Add the character range
                Array.Resize(ref ranges, ranges.Length + 1);
                ranges[ranges.Length - 1] = new charRange(pos, lineLength);
            }

            //Push the stream position back
            str.Position = position;

            //Return the buffer
            return buffer;
        }
        internal string[] getMethodNames(Stream str, out charRange[] ranges, charRange[] ignoreRanges) { 
            //Define the default outputs
            ranges = new charRange[0];

            //Buffer the stream position
            long strPos = str.Position;

            //Seek to the beginning of the stream
            str.Seek(0, SeekOrigin.Begin);

            //Define the method token
            string methodToken = "method";

            //Define the return buffer
            string[] buffer = new string[0];

            //Iterate through the stream
            while (str.Position < str.Length) { 
                //Skip comments and scopes
                while (!skipComment(str)) ;
                skipScope(str);

                //Ignore this position?
                bool ignore = false;
                foreach (charRange r in ignoreRanges) {
                    if (str.Position >= r.Start &&
                       str.Position < r.Start + r.Length) {
                           ignore = true;
                           str.Position = r.Start + r.Length;
                    }
                }
                if (ignore) { continue; }

                //Read the byte
                byte b = (byte)str.ReadByte();

                //Is it string character? (skip over)
                if (b == (byte)'"') {
                    skipString(str);
                    continue;
                }

                //Brace?
                if (b == (byte)'(') {
                    //Skip over
                    int oCount = 1;
                    int cCount = 0;
                    while (str.Position < str.Length) {
                        byte r = (byte)str.ReadByte();
                        if (r == (byte)'(') { oCount++; continue; }
                        if (r == (byte)')') {
                            cCount++;
                            if (cCount == oCount) { break; }
                            continue;
                        }
                    }
                    continue;
                }

                //Seek back
                str.Seek(-1, SeekOrigin.Current);

                //Is it a method token?
                if (Encoding.ASCII.GetString(select(str, methodToken.Length)) == methodToken) { 
                    //Define the name to build
                    string name = "";
                    str.Position += methodToken.Length;
                    
                    //Skip over the whitespaces
                    skipWhiteSpaces(str);
                    
                    //Define the character range to add
                    charRange range = new charRange();
                    range.Start = str.Position;

                    //Read the name
                    while (str.Position < str.Length) {
                        byte r = (byte)str.ReadByte();
                        if (r == (byte)' ' || r == (byte)'\n' || r == (byte)'(') { break; }
                        name += (char)r;
                    }

                    //Read the braces
                    while (str.ReadByte() != (byte)')' && str.Position < str.Length) ;
                    range.Length = str.Position - range.Start;

                    //Check the name
                    foreach (string s in buffer) {
                        if (s == name) { throw new Exception("Invalid method name \"" + name + "\" because it already exists"); }
                    }
                    foreach (IMethod m in p_Methods) {
                        if (m.Name == name) { throw new Exception("Invalid method name \"" + name + "\" because it already exists"); }
                    }
                    //Add the name to the buffer
                    Array.Resize(ref buffer, buffer.Length + 1);
                    buffer[buffer.Length - 1] = name;

                    //Add the range
                    Array.Resize(ref ranges, ranges.Length + 1);
                    ranges[ranges.Length - 1] = range;
                }

                //Seek forward
                str.Seek(1, SeekOrigin.Current);
            }

            //Push the stream back to it's previous position 
            str.Position = strPos;

            //Return the buffer
            return buffer;
        }
        internal static string getClrNameFromExternType(string t, out Type type) { 
            //Define the map for the external types            
            string[] map = { 
                "ptr=IntPtr",     
                "sbyte=SByte",
                "byte=Byte",
                "short=Int16",
                "ushort=UInt16",
                "int=Int32",
                "uint=UInt32",
                "long=Int64",
                "ulong=UInt64",
                "float=Single",
                "double=Double",
                "decimal=Decimal",
                "char=Char",
                "string=String",
                "bool=Boolean",
                "method="
            };
            Type[] typeMap = { 
                                typeof(IntPtr),
                                typeof(sbyte),
                                typeof(byte),
                                typeof(short),
                                typeof(ushort),
                                typeof(int),
                                typeof(uint),
                                typeof(long),
                                typeof(ulong),
                                typeof(float),
                                typeof(double),
                                typeof(decimal),
                                typeof(char),
                                typeof(string),
                                typeof(bool),
                                null
                             };
            
                 
            //Look for the type
            for(int c = 0; c < map.Length; c++) {
                string s = map[c];
                if (s.Split('=')[0].ToLower() == t.ToLower()) {
                    type = typeMap[c];
                    return s.Split('=')[1];
                }
            }

            //Not found
            throw new Exception("Type not found");
        }

        internal struct charRange {
            public long Start;
            public long Length;
            public charRange(long start, long length) {
                //Define the globals
                Start = start;
                Length = length;
            }
        }        
        internal enum logicOperator {
            UNKNOWN = -1,
            EQUAL = 0,
            NOT_EQUAL = 1,
            MORE_EQUAL = 2,
            LESS_EQUAL = 3,
            MORE = 4,
            LESS = 5
        }
        internal class scriptMethod : IMethod {
            private string p_Scope;
            private string p_BeginCode;
            private string p_Name;
            private string[] p_Arguments;
            private bool[] p_LogicalIgnoreMap;
            private Interpreter p_Interpreter;

            public scriptMethod(string scope, string beginCode, string name, Interpreter interpreter) { 
                //Define the globals 
                p_Scope = scope;
                p_BeginCode = beginCode;
                p_Name = name;
                p_Interpreter = interpreter;
                p_LogicalIgnoreMap = new bool[0];

                //Parse the arguments
                beginCode = beginCode.Substring(beginCode.IndexOf('(') + 1);
                beginCode = beginCode.Substring(0, beginCode.Length - 1);
                p_Arguments = new string[0];
                foreach (string s in beginCode.Split(',')) {
                    if (s.Replace(" ", "").Length == 0) { continue; }
                    Array.Resize(ref p_Arguments, p_Arguments.Length + 1);
                    Array.Resize(ref p_LogicalIgnoreMap, p_LogicalIgnoreMap.Length + 1);
                    string str = s.Replace(" ", "");
                    if (str.StartsWith("*")) {
                        str = str.Substring(1);
                        p_LogicalIgnoreMap[p_LogicalIgnoreMap.Length - 1] = true;
                    }
                    p_Arguments[p_Arguments.Length - 1] = str;
                }
            }

            public object AssignedInstance { get; set; }
            public Type[] ArgumentTypes { get { 
                //Just set every pass type to a string
                Type[] buffer = new Type[p_Arguments.Length];
                for (int c = 0; c < buffer.Length; c++) { buffer[c] = typeof(string); }
                return buffer;
            } }
            public string Name { get { return p_Name; } }
            public Type ReturnType { get { return typeof(string); } }

            public object Invoke(object[] args) { return Invoke(AssignedInstance, args); }
            public object Invoke(object instance, object[] args) {
                //Check the arguments
                if (args.Length != p_Arguments.Length) {
                    throw new Exception("Invalid arguments passed");
                }
                
                //Clone the interpreter
                Interpreter clone = (Interpreter)p_Interpreter.Clone();

                //Define the arguments as globals
                for(int c = 0; c < args.Length; c++) { 
                    //Does the name already exist?
                    foreach (IDefinition i in clone.p_Defines) {
                        if (i.Name == p_Arguments[c]) {
                            throw new Exception("Definition \"" + i.Name + "\" exists");
                        }
                    }

                    //Copy the definition?
                    if (p_LogicalIgnoreMap[c]) {
                        string aStr = args[c].ToString().Replace(" ", "");
                        bool isDef = false;
                        foreach (IDefinition d in p_Interpreter.p_Defines) {
                            if (d.Name != aStr) { continue; }
                            isDef = true;

                            //Clone the definition
                            definition cDef = new definition();
                            cDef.Name = p_Arguments[c];
                            if (d.ArrayList != null) {
                                cDef.ArrayList = d.ArrayList;
                            }
                            else if (d.DefinitionType == DefinitionType.CommonLanguageRuntimeObject) {
                                cDef.ValueInstance = d.ValueInstance;
                                cDef.setDef(DefinitionType.CommonLanguageRuntimeObject);
                            }
                            else {
                                cDef.Value = d.Value;
                                cDef.IsNumeric = d.IsNumeric;
                                cDef.IsDefaultGenerated = d.IsDefaultGenerated;                                
                            }
                            Array.Resize(ref clone.p_Defines, clone.p_Defines.Length + 1);
                            clone.p_Defines[clone.p_Defines.Length - 1] = cDef;
                            break;
                        }
                        if (isDef) { continue; }
                    }

                    //Numerical?
                    bool numeric = false;
                    try {
                        Convert.ToDouble(args[c]);
                        numeric = true;
                    }
                    catch { }

                    //Define it
                    Array.Resize(ref clone.p_Defines, clone.p_Defines.Length + 1);
                    clone.p_Defines[clone.p_Defines.Length - 1] = new definition { 
                        IsNumeric = numeric,
                        Name = p_Arguments[c],
                        Value = args[c].ToString()
                    };
                }

                //Execute the scope
                Stream scope = clone.createStrStream(p_Scope);
                RunState state = new RunState(clone, scope);
                string ret = clone.Run(scope, false, state);
          
                //Apply changes
                p_Interpreter.applyDefinitionChanges(p_Interpreter, clone);
                return ret;
            }

            public bool IgnoreLogicalArgument(int index) {
                //Just return the value from the character map
                return p_LogicalIgnoreMap[index];
            }
        }
        internal class externMethod : IMethod {
            private string p_Name, p_DllName;
            private Type p_ReturnType;
            private Type[] p_ArgumentTypes;
            private string[] p_ArgumentNameTypes;

            public externMethod(string name, string dllName, string retType, string line) { 
                //Define the globals
                p_Name = name;
                p_DllName = dllName;
                getClrNameFromExternType(retType, out p_ReturnType);
                
                //Get the arguments raw value
                string argData = line.Substring(line.IndexOf('(') + 1);
                argData = argData.Substring(0, argData.LastIndexOf(')'));
                string[] argSplit = new string[0];
                foreach (string s in argData.Split(',')) {
                    if (s.Replace(" ", "").Length == 0) { continue; }
                    Array.Resize(ref argSplit, argSplit.Length + 1);
                    argSplit[argSplit.Length - 1] = s;
                }
                p_ArgumentTypes = new Type[argSplit.Length];
                p_ArgumentNameTypes = new string[argSplit.Length];

                //Loop through the arguments
                for (int c = 0; c < p_ArgumentTypes.Length; c++) { 
                    //Grab the type
                    string aType = argSplit[c].Substring(argSplit[c].IndexOf('(') + 1);
                    aType = aType.Substring(0, aType.LastIndexOf(')'));
                    Type type = null;
                    p_ArgumentNameTypes[c] = getClrNameFromExternType(aType, out type);
                    p_ArgumentTypes[c] = type;                    
                }

                //Tell the library loader about this
                object[] args = new object[argSplit.Length];
                for (int c = 0; c < args.Length; c++) {
                    if (p_ArgumentNameTypes[c] == "IntPtr") {
                        args[c] = IntPtr.Zero;
                        continue;
                    }
                    foreach (MethodInfo m in typeof(Convert).GetMethods()) {
                        if (m.Name != "To" + p_ArgumentNameTypes[c]) { continue; }
                        try {
                            args[c] = m.Invoke(null, new object[] { 0 });
                            break;
                        }
                        catch { }
                    }
                }
                DynamicLibraryLoader.AddMethod(dllName, name, args, p_ReturnType);
            }

            public object AssignedInstance { get; set; }
            public string Name { get { return p_Name; } }
            public Type[] ArgumentTypes { get {
                return p_ArgumentTypes;
            } }
            public Type ReturnType { get { return p_ReturnType; } }

            public object Invoke(object[] args) { return Invoke(AssignedInstance, args); }
            public object Invoke(object instance, object[] args) {
                //Convert the arguments into what should be passed to the external method
                object[] eArgs = new object[args.Length];
                for (int c = 0; c < args.Length; c++) {
                    if (p_ArgumentNameTypes[c] == "IntPtr") {
                        if (args[c] == typeof(IntPtr)) { continue; }
                        try { eArgs[c] = (IntPtr)Convert.ToInt64(args[c]); }
                        catch { eArgs[c] = args[c]; }                       
                        continue;
                    }
                    foreach (MethodInfo m in typeof(Convert).GetMethods()) {
                        if (m.Name != "To" + p_ArgumentNameTypes[c]) { continue; }
                        try {
                            eArgs[c] = m.Invoke(null, new object[] { args[c] });
                            break;
                        }
                        catch { }
                    }
                }

                //Invoke the external method
                return DynamicLibraryLoader.InvokeMethod(p_DllName, p_Name, eArgs, p_ReturnType);
            }

            public bool IgnoreLogicalArgument(int index) { 
                return false;
            }
        }
    }
}