﻿using System;

namespace Plywood {
    public static class NumericOperations {
        public static ulong FromBinary(Binary binary) {
            //Define the return buffer
            ulong buffer = 0;

            //Reverse the binary
            binary.Reverse();

            //Define the current base value
            ulong baseValue = 1;

            //Loop through the binary entries
            foreach (bool b in binary) { 
                //Add the base value?
                buffer += (ulong)(b ? baseValue : 0);

                //Double the base value
                baseValue *= 2;
            }

            //Return the buffer
            return buffer;            
        }
        public static Binary ToBinary(ulong value) { 
            //Define the length and base value
            int length = 0;
            ulong baseValue = 1;
            ulong valueBuffer = 0;

            //Find the length
            for (int c = 0; c < 64; c++, length++) {
                if (valueBuffer + (ulong)baseValue >= value) { length++; break; } 
                valueBuffer += (ulong)baseValue;     
                baseValue *= 2; 
            }

            //Define the binary buffer and reset all the variables
            bool[] buffer = new bool[length];
            valueBuffer = 0;
            
            //Loop through the length
            for (int c = 0; c < length; c++) {
                //Add the appropriate binary value to the buffer
                buffer[c] = valueBuffer + (ulong)baseValue <= value;
                valueBuffer += buffer[c] ? (ulong)baseValue : 0;
                baseValue /= 2; 
            }
            
            //Return the buffer
            return buffer;
        }

        public static string ToHex(Binary binary) { 
            //Define the return buffer
            string buffer = "";

            //Get the binary string
            string binStr = binary.ToString();

            //Force binary length to be a multiple of 4
            while ((binStr.Length * 1.0f / 4).ToString().Contains(".")) {
                binStr = "0" + binStr;
            }

            //Grab the hex map
            string[] hexMap = hexBinaryMap;

            //Loop every four characters
            for (int c = 0; c < binStr.Length / 4; c++) { 
                //Get the four characters
                string sub = binStr.Substring(c * 4, 4);

                //Define if the hex has been found
                bool found = false;

                //Is the value in the hex map?
                foreach (string s in hexMap) {
                    if (s.Split('=')[1] == sub) { 
                        buffer += s.Split('=')[0];
                        found = true;
                        break;
                    }
                }

                //Has it been found
                if (found) { continue; }

                //Just add the decimal version of the binary
                buffer += FromBinary(sub).ToString();
            }

            //Return the buffer
            return buffer;
        }
        public static Binary FromHex(string hex) { 
            //Define the return buffer
            string buffer = "";

            //Grab the hex map
            string[] hexMap = hexBinaryMap;

            //Loop through every character
            foreach (char c in hex) { 
                //Define the boolean to hold if the hex value was found
                bool found = false;

                //Find the hex value in the hex map
                foreach (string s in hexMap) {
                    if (s.Split('=')[0].ToLower() == c.ToString().ToLower()) {
                        buffer += s.Split('=')[1];
                        found = true;
                        break;
                    }
                }

                //Skip?
                if (found) { continue; }

                //Attempt to convert to binary
                try {
                    Binary b = ToBinary(Convert.ToUInt64(c.ToString()));

                    //Force the binary length to be 4
                    string bStr = b;
                    while (bStr.Length != 4) { bStr = "0" + bStr; }

                    //Add the binary value to the buffer
                    buffer += bStr;
                }
                catch {
                    throw new Exception("Invalid hex");
                }
            }

            //Return the buffer
            return buffer;
        }

        public static Binary FromString(string str, bool force8bitLength) { 
            //Define the return buffer
            string buffer = "";

            //Loop through every character
            foreach (char c in str) { 
                //Grab the binary for this character
                string cB = ToBinary((ulong)c);

                //Force 8bit length?
                while (cB.Length < 8) {
                    cB = "0" + cB;
                }

                //Add to the buffer
                buffer += cB;
            }

            //Return the buffer
            return buffer;
        }

        private static string[] hexBinaryMap {
            get {
                return new string[] { 
                    "A=1010",
                    "B=1011",
                    "C=1100",
                    "D=1101",
                    "E=1110",
                    "F=1111"
                };
            }
        }
    }
}