﻿using System;
using System.IO;

namespace Plywood {
    public class TerminalInterface {
        private ITerminal[] p_Terminals;
        private RunState p_State;

        public TerminalInterface(ITerminal[] terminals, RunState state) { 
            //Define the globals
            p_Terminals = terminals;
        }

        public void Write(string value) {
            //Cause write
            sendMessage(TerminalMessageType.Write, value, false);
        }
        public void WriteLine(string line) { 
            //Cause write line
            sendMessage(TerminalMessageType.WriteLine, line, false);
        }
        public char ReadKey() { 
            //Read
            return sendMessage(TerminalMessageType.ReadKey, null, false).ToString()[0];
        }
        public string ReadLine() { 
            //Read line
            return sendMessage(TerminalMessageType.ReadLine, null, false).ToString();

        }
        public void Clear() {
            //Clear the screen
            sendMessage(TerminalMessageType.ClearScreen, null, false);
        }

        public int BufferWidth { 
            get {
                //Return the buffer height
                return (int)sendMessage(TerminalMessageType.BufferWidth, null, false);
            }
            set {
                //Set the buffer width
                sendMessage(TerminalMessageType.BufferHeight, value, true);
            }
        }
        public int BufferHeight { 
            get { 
                //Get the buffer height
                return (int)sendMessage(TerminalMessageType.BufferHeight, null, false);
            }
            set { 
                //Set the buffer height
                sendMessage(TerminalMessageType.BufferHeight, value, true);
            }
        }
        public int WindowWidth {
            get { 
                //Get the window width
                return (int)sendMessage(TerminalMessageType.WindowWidth, null, false);
            }
            set { 
                //Set the window height
                sendMessage(TerminalMessageType.WindowWidth, value, true);
            }
        }
        public int WindowHeight {
            get { 
                //Return the window height
                return (int)sendMessage(TerminalMessageType.WindowHeight, null, false);
            }
            set { 
                //Set the window height
                sendMessage(TerminalMessageType.WindowHeight, value, true);
            }
        }

        public int CursorX { 
            get { 
                //Return the cursor x
                return (int)sendMessage(TerminalMessageType.CursorX, null, false);
            }
            set { 
                //Set the cursor x
                sendMessage(TerminalMessageType.CursorX, value, true);
            }
        }
        public int CursorY {
            get { 
                //Return the cursor y
                return (int)sendMessage(TerminalMessageType.CursorY, null, false);
            }
            set { 
                //Set the cursor y
                sendMessage(TerminalMessageType.CursorY, value, true);
            }
        }

        public string Title {
            get {
                return sendMessage(TerminalMessageType.Title, null, false).ToString();
            }
            set {
                sendMessage(TerminalMessageType.Title, value, true);
            }
        }

        public bool CursorVisible {
            get { 
                //Return the cursor visibility
                return (bool)sendMessage(TerminalMessageType.CursorVisible, null, false);
            }
            set { 
                //Set the cursor visibility
                sendMessage(TerminalMessageType.CursorVisible, value, true);
            }
        }
        public int CursorSize {
            get { 
                //Return the cursor size
                return (int)sendMessage(TerminalMessageType.CursorSize, null, false);
            }
            set { 
                //Set the cursor size
                sendMessage(TerminalMessageType.CursorSize, value, true);
            }
        }

        public ConsoleColor BackColor { 
            get { 
                //Return the back color
                return (ConsoleColor)sendMessage(TerminalMessageType.BackColor, null, false);
            }
            set { 
                //Set the back color
                sendMessage(TerminalMessageType.BackColor, value, true);
            }
        }
        public ConsoleColor ForeColor {
            get { 
                //Return the fore color
                return (ConsoleColor)sendMessage(TerminalMessageType.ForeColor, null, false);
            }
            set { 
                //Set the fore color
                sendMessage(TerminalMessageType.ForeColor, value, true);
            }
        }

        public int MaximumWindowWidth { get { return (int)sendMessage(TerminalMessageType.MaximumWindowWidth, null, false); } }
        public int MaximumWindowHeight { get { return (int)sendMessage(TerminalMessageType.MaximumWindowHeight, null, false); } }

        public TextReader In { get { return new reader(this); } }
        public TextWriter Out { get { return new writer(this); } }

        private object sendMessage(TerminalMessageType type, object value, bool set) { 
            //Send message to all terminals
            foreach (ITerminal t in p_Terminals) {
                object r = t.MessageSent(p_State, type, value, set);
                if (!set) { return r; }
            }
            return null;
        }

        private class reader : TextReader {
            private TerminalInterface p_Interface;
            public reader(TerminalInterface i) { 
                //Define the globals
                p_Interface = i;
            }

            public override int Read() {
                //Read a character
                return (int)p_Interface.ReadKey();
            }
            public override string ReadLine() {
                //Read the line
                return p_Interface.ReadLine();
            }
            public override int Peek() {
                throw new NotSupportedException();
            }
            public override int Read(char[] buffer, int index, int count) {
                throw new NotSupportedException();
            }
            public override int ReadBlock(char[] buffer, int index, int count)  {
                throw new NotSupportedException();
            }
            public override string ReadToEnd() {
                throw new NotSupportedException();
            }
            public override void Close() {
                throw new NotSupportedException();
            }            
        }
        private class writer : TextWriter {
            private TerminalInterface p_Interface;
            public writer(TerminalInterface i) { 
                //Define the globals
                p_Interface = i;
            }

            public override void Write(bool value) { p_Interface.Write(value.ToString()); }
            public override void Write(char[] buffer) {
                foreach (char c in buffer) {
                    p_Interface.Write(c.ToString());
                }
            }
            public override void Write(char value) { p_Interface.Write(value.ToString()); }
            public override void Write(char[] buffer, int index, int count) {
                throw new NotSupportedException();
            }
            public override void Write(decimal value) { p_Interface.Write(value.ToString()); }
            public override void Write(double value) { p_Interface.Write(value.ToString()); }
            public override void Write(float value) { p_Interface.Write(value.ToString()); }
            public override void Write(int value) { p_Interface.Write(value.ToString()); }
            public override void Write(long value) { p_Interface.Write(value.ToString()); }
            public override void Write(object value) { p_Interface.Write(value.ToString()); }
            public override void Write(string format, object arg0) {
                throw new NotSupportedException();
            }
            public override void Write(string format, object arg0, object arg1) {
                throw new NotSupportedException();
            }
            public override void Write(string format, object arg0, object arg1, object arg2) {
                throw new NotSupportedException();
            }
            public override void Write(string format, params object[] arg) {
                throw new NotSupportedException();
            }
            public override void Write(string value) { p_Interface.Write(value); }
            public override void Write(uint value) { p_Interface.Write(value.ToString()); }
            public override void Write(ulong value) { p_Interface.Write(value.ToString()); }
            public override void WriteLine() { WriteLine(""); }
            public override void WriteLine(string value) { p_Interface.WriteLine(value); }
            public override void WriteLine(bool value) { WriteLine(value.ToString()); }
            public override void WriteLine(char value) { WriteLine(value.ToString()); }
            public override void WriteLine(char[] buffer) {
                Write(buffer);
                WriteLine();
            }
            public override void WriteLine(char[] buffer, int index, int count) {
                throw new NotSupportedException();
            }
            public override void WriteLine(decimal value) { WriteLine(value.ToString()); }
            public override void WriteLine(double value) { WriteLine(value.ToString()); }
            public override void WriteLine(float value) { WriteLine(value.ToString()); }
            public override void WriteLine(int value) { WriteLine(value.ToString()); }
            public override void WriteLine(long value) { WriteLine(value.ToString()); }
            public override void WriteLine(object value) { WriteLine(value.ToString()); }
            public override void WriteLine(string format, object arg0)  {
                throw new NotSupportedException();
            }
            public override void WriteLine(string format, object arg0, object arg1) {
                throw new NotSupportedException();
            }
            public override void WriteLine(string format, object arg0, object arg1, object arg2) {
                throw new NotSupportedException();
            }
            public override void WriteLine(string format, params object[] arg) {
                throw new NotSupportedException();
            }
            public override void WriteLine(uint value) { WriteLine(value.ToString()); }
            public override void WriteLine(ulong value) { WriteLine(value.ToString()); }
            public override void Flush() {
                throw new NotSupportedException();
            }
            public override IFormatProvider FormatProvider {
                get {
                    throw new NotSupportedException();
                }
            }
            public override string NewLine {
                get {
                    return "\n";
                }
                set {
                    throw new NotSupportedException();
                }
            }
            public override void Close() {
                throw new NotSupportedException();
            }
         
            public override System.Text.Encoding Encoding { get { return System.Text.Encoding.ASCII; } }
        }
    }
}