﻿using System;
using System.IO;

namespace Plywood {
    public class WindowsTerminal : ITerminal {
        public object MessageSent(RunState state, TerminalMessageType type, object arg, bool set) {
            //Valid?
            if (type == TerminalMessageType.Unknown) {
                throw new Exception("Type cannot be unknown");
            }

            //WriteLine? if so, simulate a write
            if (type == TerminalMessageType.WriteLine) {
                MessageSent(state, TerminalMessageType.Write, arg.ToString() + "\n", false);
                return null;
            }
            
            //Buffer width/height?
            if (type == TerminalMessageType.BufferWidth) {
                if (set) { Console.BufferWidth = (int)arg; return null; }
                return Console.BufferWidth;
            }
            else if (type == TerminalMessageType.BufferHeight) {
                if (set) { Console.BufferHeight = (int)arg; return null; }
                return Console.BufferHeight;
            }
            //Window width/height?
            else if (type == TerminalMessageType.WindowWidth) {
                if (set) { Console.WindowWidth = (int)arg; return null; }
                return Console.WindowWidth;
            }
            else if (type == TerminalMessageType.WindowHeight) {
                if (set) { Console.WindowHeight = (int)arg; return null; }
                return Console.WindowHeight;
            }
            //Window maximum size?
            else if (type == TerminalMessageType.MaximumWindowWidth)  {
                return Console.LargestWindowWidth;
            }
            else if (type == TerminalMessageType.MaximumWindowHeight) {
                return Console.LargestWindowHeight;
            }
            //Cursor position?
            else if (type == TerminalMessageType.CursorX) {
                if (set) { Console.CursorLeft = (int)arg; return null; }
                return Console.CursorLeft;
            }
            else if (type == TerminalMessageType.CursorY)  {
                if (set) { Console.CursorTop = (int)arg; return null; }
                return Console.CursorTop;
            }
            //Console color?
            else if (type == TerminalMessageType.BackColor) {
                if (set) { Console.BackgroundColor = (ConsoleColor)arg; }
                return Console.BackgroundColor;
            }
            //Fore color?
            else if (type == TerminalMessageType.ForeColor) {
                if (set) { Console.ForegroundColor = (ConsoleColor)arg; }
                return Console.ForegroundColor;
            }


            //Switch to the type to handle output operations
            switch (type) {
                //Clear screen?
                case TerminalMessageType.ClearScreen:
                    Console.Clear();

                    try {
                        state.p_PrintStreamBuffer.Close();
                        state.p_PrintStreamBuffer = new FileStream(state.p_PrintStringFile, FileMode.Create);
                    }
                    catch { }
                    return null;
                //Write?
                case TerminalMessageType.Write:
                    Console.Write((string)arg);
                    try { state.p_PrintStreamBuffer.Write(System.Text.Encoding.ASCII.GetBytes((string)arg), 0, arg.ToString().Length); }
                    catch { }
                    return null;
                //Read key?
                case TerminalMessageType.ReadKey:
                    char ch = Console.ReadKey().KeyChar;
                    state.p_PrintStreamBuffer.WriteByte((byte)ch);
                    return ch;
                //Read line?
                case TerminalMessageType.ReadLine:
                    string r = Console.ReadLine() + "\n";
                    state.p_PrintStreamBuffer.Write(System.Text.Encoding.ASCII.GetBytes(r), 0, r.Length);
                    return r.Substring(0, r.Length - 1);
                //Title?
                case TerminalMessageType.Title:
                    if (set) { Console.Title = (string)arg; }
                    return Console.Title;
                //Cursor size?
                case TerminalMessageType.CursorSize:
                    if (set) { Console.CursorSize = (int)arg; }
                    return Console.CursorSize;
                //Cursor visible?
                case TerminalMessageType.CursorVisible:
                    if (set) { Console.CursorVisible = (bool)arg; }
                    return Console.CursorVisible;
                default:
                    return null;
            }
        }

        public bool Enabled { get; set; }
    }
}